"""
Radiative heating of contrail cirrus.

Contrails absorb incoming solar radiation and outgoing longwave radiation, causing it to heat up.

1. The additional heating energy drives a local updraft
   (but this is negligible and not included in CoCiP);
2. The  differential heating rate drives local turbulence.
   As radiative emission from the contrail is limited by its low temperature,
   the net result is deposition of radiative energy in the contrail.

These equations are provided by Ulrich Schumann and not included in any scientific publications,
The scientific rationale of including these effects can be found in the references below:

References
----------
- :cite:`jensenSpreadingGrowthContrails1998`
- :cite:`schumannLifeCycleIndividual2017`
"""

from __future__ import annotations

from dataclasses import dataclass

import numpy as np
import numpy.typing as npt

from pycontrails.physics import constants


@dataclass
class RadiativeHeatingConstants:
    """Constants/coefficients used to calculate the radiative heating rate.

    Constants/coefficients used to calculate the:
     - shortwave differential heating rate
     - longwave differential heating rate
     - shortwave heating rate
     - longwave heating rate

    These coefficients were calibrated based on a least-squares fit relative
    to the libRadtran radiative transfer model. (Ulrich Schumann, personal communication).

    References
    ----------
    - :cite:`schumannRadiativeHeatingContrail2010`
    """

    #: Coefficients for shortwave differential heating rate
    dacth: float = 0.205747e01
    dacth3: float = 0.898366e00
    dbcth: float = 0.791045e00
    dccth: float = 0.612725e00
    ddcth: float = 0.517342e-02
    dexalb: float = 0.267568e01
    dfrsw: float = 0.139286e01
    dgalbs: float = 0.178497e01
    d_gamma: float = 0.142104e01
    d_gamma_s: float = 0.882497e00
    dqsw: float = 0.631427e-01
    draddsw: float = 0.261780e00
    dtt: float = 0.339171e-01

    #: Coefficients for longwave differential heating rate
    dak: float = 0.357181e01
    dcrhi: float = 0.623019e-01
    ddelta: float = 0.198000e01
    dfrlw: float = 0.609262e00
    dqlw: float = 0.100000e-05
    dqrlw: float = 0.160286e00
    draddlw: float = 0.898529e-05
    dsigma: float = 0.159884e-06

    #: Coefficients for shortwave heating rate
    acth: float = 0.156899e01
    bcth: float = 0.875130e00
    ccth: float = 0.112445e01
    dcth: float = 0.236688e-01
    exal_b: float = 0.410705e00
    fr_sw: float = 0.537577e01
    gamma_r: float = 0.762254e00
    q_sw: float = 0.454176e-01
    radd_sw: float = 0.991554e00
    ttt: float = 0.985031e-01

    #: Coefficients for longwave heating rate
    ak: float = 0.294930e01
    crhi: float = 0.174422e00
    czlw: float = 0.393884e-01
    delta: float = 0.860746e00
    fr_lw: float = 0.760423e00
    q_lw: float = 0.152075e02
    radd_lw: float = 0.308486e-02
    sigma: float = 0.253499e-04


RAD_HEAT = RadiativeHeatingConstants()


def convective_velocity_scale(
    depth_eff: npt.NDArray[np.float_],
    eff_heat_rate: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Calculate the convective velocity scale, i.e., vertical mixing rate.

    Parameters
    ----------
    depth_eff : npt.NDArray[np.float_]
        Effective depth of the contrail plume, [:math:`m`]
    eff_heat_rate: npt.NDArray[np.float_]
        Effective heating rate, i.e., rate of which the contrail plume is heated, [:math:`K s^{-1}`]
    air_temperature : npt.NDArray[np.float_]
        Ambient temperature for each waypoint, [:math:`K`]

    Returns
    -------
    npt.NDArray[np.float_]
        Convective velocity scale, [:math:`m s^{-1}`]
    """
    return ((constants.g * depth_eff**2 * np.maximum(-eff_heat_rate, 0)) / air_temperature) ** (
        1 / 3
    )


def effective_heating_rate(
    d_heat_rate: npt.NDArray[np.float_],
    cumul_rad_heat: npt.NDArray[np.float_],
    dT_dz: npt.NDArray[np.float_],
    depth: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """Calculate effective heating rate.

    The effective heating rate accounts for the heat required to overcome the stable stratification.
    Turbulence will occur after the cumulative heating overcomes the stable stratification.

    Parameters
    ----------
    d_heat_rate: npt.NDArray[np.float_]
        Differential heating rate, i.e., rate of which the contrail
        plume is heated, [:math:`K s^{-1}`]
    cumul_rad_heat: npt.NDArray[np.float_]
        Cumulative solar and terrestrial radiative heating energy
        absorbed by the contrail, [:math:`K`]
    dT_dz: npt.NDArray[np.float_]
        Temperature gradient with respect to altitude (dz), [:math:`K m^{-1}`]
    depth : npt.NDArray[np.float_]
        Contrail depth at each waypoint, [:math:`m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Effective heating rate, [:math:`K s^{-1}`]
    """
    filt = cumul_rad_heat > 0.0
    heat_denom = 0.5 * dT_dz[filt] * depth[filt]
    heat_denom.clip(min=0.0, out=heat_denom)

    heat_ratio = np.zeros_like(cumul_rad_heat)
    heat_ratio[filt] = cumul_rad_heat[filt] / (cumul_rad_heat[filt] + heat_denom)

    return d_heat_rate * heat_ratio


def differential_heating_rate(
    air_temperature: npt.NDArray[np.float_],
    rhi: npt.NDArray[np.float_],
    rho_air: npt.NDArray[np.float_],
    r_ice_vol: npt.NDArray[np.float_],
    depth_eff: npt.NDArray[np.float_],
    tau_contrail: npt.NDArray[np.float_],
    tau_cirrus: npt.NDArray[np.float_],
    sd0: npt.NDArray[np.float_],
    sdr: npt.NDArray[np.float_],
    rsr: npt.NDArray[np.float_],
    olr: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Calculate the differential heating rate affecting the contrail plume.

    Differential heating rate is the heating rate difference between the upper and
    lower half of the cirrus layer. The radiative heating effect is dominated by the
    longwave component. Therefore, this output will always be a positive
    value (i.e., warmer at the upper contrail edge and cooler at the bottom).

    Parameters
    ----------
    air_temperature : npt.NDArray[np.float_]
        Ambient temperature at each waypoint, [:math:`K`]
    rhi : npt.NDArray[np.float_]
        Relative humidity with respect to ice at each waypoint
    rho_air : npt.NDArray[np.float_]
        Density of air for each waypoint, [:math:`kg m^{-3}`]
    r_ice_vol : npt.NDArray[np.float_]
        Ice particle volume mean radius, [:math:`m`]
    depth_eff : npt.NDArray[np.float_]
        Effective depth of the contrail plume, [:math:`m`]
    tau_contrail : npt.NDArray[np.float_]
        Contrail optical depth for each waypoint
    tau_cirrus : npt.NDArray[np.float_]
        Optical depth of numerical weather prediction (NWP) cirrus above the contrail
    sd0 : npt.NDArray[np.float_]
        Solar constant, [:math:`W m^{-2}`]
    sdr : npt.NDArray[np.float_]
        Solar direct radiation, [:math:`W m^{-2}`]
    rsr : npt.NDArray[np.float_]
        Reflected solar radiation, [:math:`W m^{-2}`]
    olr : npt.NDArray[np.float_]
        Outgoing longwave radiation at each waypoint, [:math:`W m^{-2}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Differential heating rate, [:math:`K s^{-1}`]
    """
    r_ice_vol_um = r_ice_vol * 1e6
    cp_contrail = contrail_heat_capacity(rho_air, depth_eff)
    d_heat_rate_sw = differential_heating_rate_shortwave(
        cp_contrail, r_ice_vol_um, tau_contrail, tau_cirrus, sd0, sdr, rsr
    )
    d_heat_rate_lw = differential_heating_rate_longwave(
        air_temperature, rhi, cp_contrail, r_ice_vol_um, tau_contrail, tau_cirrus, olr
    )
    return np.minimum(d_heat_rate_sw + d_heat_rate_lw, 0.0)


def differential_heating_rate_shortwave(
    cp_contrail: npt.NDArray[np.float_],
    r_ice_vol_um: npt.NDArray[np.float_],
    tau_contrail: npt.NDArray[np.float_],
    tau_cirrus: npt.NDArray[np.float_],
    sd0: npt.NDArray[np.float_],
    sdr: npt.NDArray[np.float_],
    rsr: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    r"""
    Calculate shortwave differential heating rate.

    Incoming solar radiation heats the contrail top. Therefore, this output will
    always be a positive value (i.e., warmer at the upper contrail edge and cooler
    at the bottom). This effect is approximated using a parametric model
    that is calibrated with the libRadtran radiative transfer model.

    Parameters
    ----------
    cp_contrail : npt.NDArray[np.float_]
        Contrail heat capacity per unit length and width, [:math:`J K^{-1} m^{-2}`]
    r_ice_vol_um : npt.NDArray[np.float_]
        Ice particle volume mean radius, [:math:`\mu m`]
    tau_contrail : npt.NDArray[np.float_]
        Contrail optical depth for each waypoint
    tau_cirrus : npt.NDArray[np.float_]
        Optical depth of numerical weather prediction (NWP) cirrus above the contrail
    sd0 : npt.NDArray[np.float_]
        Solar constant, [:math:`W m^{-2}`]
    sdr : npt.NDArray[np.float_]
        Solar direct radiation, [:math:`W m^{-2}`]
    rsr : npt.NDArray[np.float_]
        Reflected solar radiation, [:math:`W m^{-2}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Shortwave component of the differential heating rate, [:math:`K s^{-1}`]
    """
    # short circuit if no waypoints have sdr > 0
    if not np.any(sdr > 0):
        return np.zeros_like(sdr)

    mue = np.minimum(sdr / sd0, 1.0)
    tau_eff = tau_contrail / (mue + 1.0e-6)

    return (
        (1 - np.exp(-RAD_HEAT.dqsw * r_ice_vol_um))
        * (RAD_HEAT.dtt * sdr - RAD_HEAT.ddcth * rsr)
        * tau_contrail
        * (-RAD_HEAT.d_gamma_s * tau_contrail + tau_eff)
        * (1 / cp_contrail)
        * (
            1
            - RAD_HEAT.dacth * mue
            + RAD_HEAT.d_gamma * mue**2
            + (RAD_HEAT.dacth3 - 1) * mue**3
        )
        * np.exp(RAD_HEAT.dbcth * tau_cirrus - RAD_HEAT.dccth * tau_cirrus / (mue + 1e-6))
        * np.exp(-RAD_HEAT.dgalbs * tau_contrail * (1 - mue) ** RAD_HEAT.dexalb)
        * mue**RAD_HEAT.draddsw
        * (1 + RAD_HEAT.dfrsw * (1 - mue) ** 2)
    )


def differential_heating_rate_longwave(
    air_temperature: npt.NDArray[np.float_],
    rhi: npt.NDArray[np.float_],
    cp_contrail: npt.NDArray[np.float_],
    r_ice_vol_um: npt.NDArray[np.float_],
    tau_contrail: npt.NDArray[np.float_],
    tau_cirrus: npt.NDArray[np.float_],
    olr: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    r"""
    Calculate longwave differential heating rate.

    Contrails absorb outgoing longwave radiation emitted from the warm surface
    below and heats it up. Therefore, this output will always be a negative value
    (i.e., warmer at the lower contrail edge and cooler at the top). This effect
    is approximated using a parametric model that is calibrated with the
    libRadtran radiative transfer model.

    Parameters
    ----------
    air_temperature : npt.NDArray[np.float_]
        Ambient temperature at each waypoint, [:math:`K`]
    rhi : npt.NDArray[np.float_]
        Relative humidity with respect to ice at each waypoint
    cp_contrail : npt.NDArray[np.float_]
        Contrail heat capacity per unit length and width, [:math:`J K^{-1} m^{-2}`]
    r_ice_vol_um : npt.NDArray[np.float_]
        Ice particle volume mean radius, [:math:`\mu m`]
    tau_contrail : npt.NDArray[np.float_]
        Contrail optical depth for each waypoint
    tau_cirrus : npt.NDArray[np.float_]
        Optical depth of numerical weather prediction (NWP) cirrus above the contrail
    olr : npt.NDArray[np.float_]
        Outgoing longwave radiation at each waypoint, [:math:`W m^{-2}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Longwave component of the differential heating rate, [:math:`K s^{-1}`]
    """
    cool = RAD_HEAT.dsigma * air_temperature**RAD_HEAT.dak
    epsc = 1 - np.exp(-RAD_HEAT.ddelta * (tau_contrail + tau_cirrus))
    return (
        -RAD_HEAT.dfrlw
        * (1 / cp_contrail)
        * (olr - cool)
        * (epsc / RAD_HEAT.ddelta)
        * tau_contrail
        * np.exp(-RAD_HEAT.dqlw * tau_cirrus)
        * np.maximum(1 - RAD_HEAT.draddlw * 10 / (r_ice_vol_um + 30), 0.0)
        * (1 - np.exp(-RAD_HEAT.dqrlw * r_ice_vol_um))
        * np.exp(-(rhi - 0.9) * RAD_HEAT.dcrhi)
    )


def heating_rate(
    air_temperature: npt.NDArray[np.float_],
    rhi: npt.NDArray[np.float_],
    rho_air: npt.NDArray[np.float_],
    r_ice_vol: npt.NDArray[np.float_],
    depth_eff: npt.NDArray[np.float_],
    tau_contrail: npt.NDArray[np.float_],
    tau_cirrus: npt.NDArray[np.float_],
    sd0: npt.NDArray[np.float_],
    sdr: npt.NDArray[np.float_],
    rsr: npt.NDArray[np.float_],
    olr: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Calculate the heating rate affecting the contrail plume.

    This is the average heating rate over the contrail cirrus layer.

    Parameters
    ----------
    air_temperature : npt.NDArray[np.float_]
        Ambient temperature at each waypoint, [:math:`K`]
    rhi : npt.NDArray[np.float_]
        Relative humidity with respect to ice at each waypoint
    rho_air : npt.NDArray[np.float_]
        Density of air for each waypoint, [:math:`kg m^{-3}`]
    r_ice_vol : npt.NDArray[np.float_]
        Ice particle volume mean radius, [:math:`m`]
    depth_eff : npt.NDArray[np.float_]
        Effective depth of the contrail plume, [:math:`m`]
    tau_contrail : npt.NDArray[np.float_]
        Contrail optical depth for each waypoint
    tau_cirrus : npt.NDArray[np.float_]
        Optical depth of numerical weather prediction (NWP) cirrus above the contrail
    sd0 : npt.NDArray[np.float_]
        Solar constant, [:math:`W m^{-2}`]
    sdr : npt.NDArray[np.float_]
        Solar direct radiation, [:math:`W m^{-2}`]
    rsr : npt.NDArray[np.float_]
        Reflected solar radiation, [:math:`W m^{-2}`]
    olr : npt.NDArray[np.float_]
        Outgoing longwave radiation at each waypoint, [:math:`W m^{-2}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Heating rate, [:math:`K s^{-1}`]
    """
    r_ice_vol_um = r_ice_vol * 1e6
    cp_contrail = contrail_heat_capacity(rho_air, depth_eff)
    heat_rate_sw = heating_rate_shortwave(
        cp_contrail, r_ice_vol_um, tau_contrail, tau_cirrus, sd0, sdr, rsr
    )
    heat_rate_lw = heating_rate_longwave(
        air_temperature, rhi, cp_contrail, r_ice_vol_um, tau_contrail, tau_cirrus, olr
    )
    return heat_rate_sw + heat_rate_lw


def heating_rate_shortwave(
    cp_contrail: npt.NDArray[np.float_],
    r_ice_vol_um: npt.NDArray[np.float_],
    tau_contrail: npt.NDArray[np.float_],
    tau_cirrus: npt.NDArray[np.float_],
    sd0: npt.NDArray[np.float_],
    sdr: npt.NDArray[np.float_],
    rsr: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    r"""Calculate shortwave heating rate.

    Parameters
    ----------
    cp_contrail : npt.NDArray[np.float_]
        Contrail heat capacity per unit length and width, [:math:`J K^{-1} m^{-2}`]
    r_ice_vol_um : npt.NDArray[np.float_]
        Ice particle volume mean radius, [:math:`\mu m`]
    tau_contrail : npt.NDArray[np.float_]
        Contrail optical depth for each waypoint
    tau_cirrus : npt.NDArray[np.float_]
        Optical depth of numerical weather prediction (NWP) cirrus above the contrail
    sd0 : npt.NDArray[np.float_]
        Solar constant, [:math:`W m^{-2}`]
    sdr : npt.NDArray[np.float_]
        Solar direct radiation, [:math:`W m^{-2}`]
    rsr : npt.NDArray[np.float_]
        Reflected solar radiation, [:math:`W m^{-2}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Shortwave component of heating rate, [:math:`K s^{-1}`]
    """
    # short circuit if no waypoints have sdr > 0
    if not np.any(sdr > 0):
        return np.zeros_like(sdr)

    mue = np.minimum(sdr / sd0, 1.0)
    tau_eff = tau_contrail / (mue + 1.0e-6)
    heat_rate_sw = (
        (1 - np.exp(-RAD_HEAT.q_sw * r_ice_vol_um))
        * (RAD_HEAT.ttt * sdr + RAD_HEAT.dcth * rsr)
        * tau_eff
        * (1 / cp_contrail)
        * (1 - RAD_HEAT.acth * mue + RAD_HEAT.gamma_r * mue**2)
        * np.exp(RAD_HEAT.bcth * tau_cirrus - RAD_HEAT.ccth * tau_cirrus / (mue + 1e-6))
        * np.exp(-tau_contrail * (1 - mue) ** 2)
        * mue
    )
    return np.maximum(heat_rate_sw, 0)


def heating_rate_longwave(
    air_temperature: npt.NDArray[np.float_],
    rhi: npt.NDArray[np.float_],
    cp_contrail: npt.NDArray[np.float_],
    r_ice_vol_um: npt.NDArray[np.float_],
    tau_contrail: npt.NDArray[np.float_],
    tau_cirrus: npt.NDArray[np.float_],
    olr: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    r"""Calculate longwave heating rate.

    Parameters
    ----------
    air_temperature : npt.NDArray[np.float_]
        Ambient temperature at each waypoint, [:math:`K`]
    rhi : npt.NDArray[np.float_]
        Relative humidity with respect to ice at each waypoint
    cp_contrail : npt.NDArray[np.float_]
        Contrail heat capacity per unit length and width, [:math:`J K^{-1} m^{-2}`]
    r_ice_vol_um : npt.NDArray[np.float_]
        Ice particle volume mean radius, [:math:`\mu m`]
    tau_contrail : npt.NDArray[np.float_]
        Contrail optical depth for each waypoint
    tau_cirrus : npt.NDArray[np.float_]
        Optical depth of numerical weather prediction (NWP) cirrus above the contrail
    olr : npt.NDArray[np.float_]
        Outgoing longwave radiation at each waypoint, [:math:`W m^{-2}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Longwave component of heating rate, [:math:`K s^{-1}`]
    """
    fzlw = np.exp(-(rhi - 0.9) * RAD_HEAT.czlw)
    cool = RAD_HEAT.sigma * air_temperature**RAD_HEAT.ak
    epsc = 1 - np.exp(-RAD_HEAT.delta * (tau_contrail + tau_cirrus))
    heat_rate_lw = (
        np.exp(-(rhi - 0.9) * RAD_HEAT.crhi)
        * RAD_HEAT.fr_lw
        * (1 / cp_contrail)
        * (olr / fzlw - cool * fzlw)
        * (epsc / RAD_HEAT.delta)
        * np.exp(-RAD_HEAT.q_lw * tau_cirrus)
        * (1 + RAD_HEAT.radd_lw / (r_ice_vol_um + 30))
    )
    return np.maximum(heat_rate_lw, 0)


def contrail_heat_capacity(
    rho_air: npt.NDArray[np.float_], depth_eff: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate contrail heat capacity per unit length and width.

    Parameters
    ----------
    rho_air : npt.NDArray[np.float_]
        density of air for each waypoint, [:math:`kg m^{-3}`]
    depth_eff: npt.NDArray[np.float_]
        Effective depth of the contrail plume, [:math:`m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Contrail heat capacity per unit length and width, [:math:`J K^{-1} m^{-2}`]
    """
    return depth_eff * rho_air * constants.c_pd



"""Ensure parity between Cocip and CocipGrid models."""

from __future__ import annotations

import numpy as np
import pytest

try:
    from pycontrails.models.cocipgrid import CocipGrid
except ImportError:
    pytest.skip("CocipGrid not available", allow_module_level=True)

from pycontrails import Flight, MetDataset
from pycontrails.models.cocip import Cocip
from pycontrails.models.humidity_scaling import ExponentialBoostHumidityScaling
from pycontrails.utils.synthetic_flight import SyntheticFlight
from tests import BADA4_PATH

# Found by running big for-loop over many seeds
# Each of these seeds creates a contrail-producing flight
# Only running on a subset of seeds to keep tests short
seeds = [35, 51, 184, 257, 324, 365, 409]
# 414  # this seed actually fails! The ones below succeed.
# 423
# 479
# 533
# 593
# 658
# 701
# 782
# 796
# 806
# 849
# 851
# 914
# 957
# 960


@pytest.fixture(scope="module")
def met(met_cocip1_module_scope: MetDataset):
    """Namespace convenience."""
    return met_cocip1_module_scope


@pytest.fixture(scope="module")
def rad(rad_cocip1_module_scope: MetDataset):
    """Namespace convenience."""
    return rad_cocip1_module_scope


@pytest.fixture(params=seeds)
def fl(met: MetDataset, request) -> Flight:
    """Create synthetic flight bounded by `met`."""
    # cocip_met has 13 time slices
    assert met["time"].size == 13

    # Explicitly slice longitude and latitude in order to keep trajectory in bounds
    # when running Cocip. In this test, we set interpolation_bounds_error=True
    bounds: dict[str, np.ndarray] = {
        "longitude": met["longitude"].values[1:-1],
        "latitude": met["latitude"].values[1:-1],
        "level": np.array([250]),
        "time": np.array([met["time"].values[0], met["time"].values[6]]),
    }

    syn = SyntheticFlight(
        bounds,
        "A320",
        u_wind=met["eastward_wind"],
        v_wind=met["northward_wind"],
        seed=request.param,
        bada4_path=BADA4_PATH,
    )
    return syn()


@pytest.mark.skipif(not BADA4_PATH.is_dir(), reason="BADA4 not available")
def test_parity(fl: Flight, met: MetDataset, rad: MetDataset):
    """Ensure substantial parity between `Cocip` and `CocipGrid`."""

    fl["azimuth"] = fl.segment_azimuth()

    # Run BADA up front
    from pycontrails.ext.bada import BADAFlight

    bada_model = BADAFlight(met=met, bada4_path=BADA4_PATH)
    fl = bada_model.eval(fl)

    # Confirm that fl has no additional variables
    assert len(fl.data) == 16

    model_params = {
        "dt_integration": np.timedelta64(2, "m"),  # keep small to ensure parity
        "max_age": np.timedelta64(1, "h"),
        "interpolation_bounds_error": True,
        "humidity_scaling": ExponentialBoostHumidityScaling(),
    }
    cocip = Cocip(met=met, rad=rad, **model_params, aircraft_performance=bada_model)
    out1 = cocip.eval(fl)

    # source data has been copied on eval
    assert len(fl.data) == 16
    assert len(cocip.source.data) == 68
    downwash1 = cocip._downwash_contrail.dataframe.set_index("waypoint")

    cg = CocipGrid(
        met=met,
        rad=rad,
        **model_params,
        bada4_path=BADA4_PATH,
        verbose_outputs_evolution=True,
        verbose_outputs_formation=True,
    )
    out2 = cg.eval(fl)
    assert len(fl.data) == 16
    assert len(cg.source.data) == 25

    # Clunky but reliable way to access CocipGrid downwash data
    downwash2 = cg.contrail.groupby("index").first()

    assert len(downwash1) == len(downwash2)
    np.testing.assert_array_equal(downwash1.index, downwash2.index)

    # ------------------------------------------------------------
    # Confirm exact agreement on verbose formation outputs
    # ------------------------------------------------------------

    # Ignore anything fl came with -- those are already identical
    common_keys = set(cg.source).intersection(cocip.source).difference(fl)
    assert common_keys == {"contrail_age", "nvpm_ei_n", "sac", "specific_humidity", "rhi"}
    exclude_keys = ["contrail_age"]  # This is checked below
    for key in common_keys.difference(exclude_keys):
        np.testing.assert_array_equal(cocip.source[key], cg.source[key], err_msg=key)

    # ---------------------------------------------------------
    # Confirm very tight agreement between initially persistent
    # ---------------------------------------------------------

    common_keys = set(downwash1).intersection(downwash2)
    assert len(common_keys) == 43
    exclude_keys = ["segment_length", "dsn_dz", "time", "formation_time"]

    for key in common_keys.difference(exclude_keys):
        np.testing.assert_array_equal(downwash1[key], downwash2[key], err_msg=key)

    different_keys = [
        ("vertical_velocity", "lagrangian_tendency_of_air_pressure"),
        ("u_wind", "eastward_wind"),
        ("u_wind_lower", "eastward_wind_lower"),
        ("v_wind", "northward_wind"),
        ("v_wind_lower", "northward_wind_lower"),
    ]
    for key1, key2 in different_keys:
        msg = f"{key1} and {key2} do not agree"
        np.testing.assert_array_equal(downwash1[key1], downwash2[key2], err_msg=msg)

    # The segment_length variable is fundamentally different
    # And dsn_dz is different as a consequence of the Cocip continuity conventions
    assert np.all(downwash2["segment_length"] == cg.params["segment_length"])
    continuous = downwash1["continuous"]
    np.testing.assert_allclose(
        downwash1["dsn_dz"][continuous],
        downwash2["dsn_dz"][continuous],
        atol=1e-6,
    )
    assert np.all(downwash1["dsn_dz"][~continuous] == 0)
    assert np.all(downwash2["dsn_dz"][~continuous] != 0)

    # No negative EF
    assert not np.any(out1["ef"] < 0)
    assert not np.any(out2["ef_per_m"] < 0)

    # -------------------------------------------------
    # Confirm general agreement between model EF output
    # -------------------------------------------------

    # Convert ef_per_m -> ef_per_waypoint
    ef_per_waypoint = out2["ef_per_m"] * out1["segment_length"]

    # Ignore waypoints with small EF -- these often arise from continuity conventions
    filt = out1["ef"] > 1e10
    assert np.sum(filt) > 5  # at least 5 waypoints with substantial EF

    # Apart from these continuity issues, close agreement
    # This assertion is critical ... it is the glue holding together Cocip and CocipGrid
    # The rtol cannot be improved much beyond the threshold below
    # Primary differences are due to continuity conventions
    # Secondary differences are due to different specific_humidity values between the two models
    ef1 = out1["ef"][filt]
    ef2 = ef_per_waypoint[filt]
    np.testing.assert_allclose(ef1, ef2, rtol=0.1)

    # ----------------------------------------------------
    # Confirm general agreement between model contrail age
    # ----------------------------------------------------

    # Borrow the filter from above
    age1 = out1["contrail_age"][filt]
    age2 = out2["contrail_age"][filt]
    np.testing.assert_array_equal(age1, age2, err_msg="contrail_age")



"""Jet aircraft trajectory and performance parameters.

This module includes common functions to calculate jet aircraft trajectory
and performance parameters, including fuel quantities, mass, thrust setting
and propulsion efficiency.
"""

from __future__ import annotations

import logging

import numpy as np
import numpy.typing as npt

from pycontrails.core import flight
from pycontrails.physics import constants, units
from pycontrails.utils.types import ArrayScalarLike

logger = logging.getLogger(__name__)


# -------------------
# Aircraft performance
# -------------------


def clip_mach_number(
    true_airspeed: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
    max_mach_number: float,
) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]:
    r"""Compute the Mach number from the true airspeed and ambient temperature.

    This method clips the computed Mach number to the value of `max_mach_number`.

    Parameters
    ----------
    true_airspeed : npt.NDArray[np.float_]
        Array of true airspeed, [:math:`m \ s^{-1}`]
    air_temperature : npt.NDArray[np.float_]
        Array of ambient temperature, [:math: `K`]
    max_mach_number : float
        Maximum mach number associated to aircraft, [:math: `Ma`]. If no clipping
        is desired, this can be set tp `np.inf`.

    Returns
    -------
    tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]] :
        Pair of true airspeed and Mach number arrays. Both are corrected so that
        the Mach numbers are clipped at `max_mach_number`.
    """
    mach_num = units.tas_to_mach_number(true_airspeed, air_temperature)

    is_unrealistic = mach_num > max_mach_number
    if np.any(is_unrealistic):
        msg = (
            f"Unrealistic Mach numbers found. Discovered {np.sum(is_unrealistic)} / "
            f"{is_unrealistic.size} values exceeding this, the largest of which "
            f"is {np.nanmax(mach_num)}. These are all clipped at {max_mach_number}."
        )
        logger.debug(msg)

    max_tas = units.mach_number_to_tas(max_mach_number, air_temperature)
    adjusted_mach_num = np.where(is_unrealistic, max_mach_number, mach_num)
    adjusted_true_airspeed = np.where(is_unrealistic, max_tas, true_airspeed)

    return adjusted_true_airspeed, adjusted_mach_num


def overall_propulsion_efficiency(
    true_airspeed: npt.NDArray[np.float_],
    F_thrust: npt.NDArray[np.float_],
    fuel_flow: npt.NDArray[np.float_],
    q_fuel: float,
    is_descent: npt.NDArray[np.bool_] | None,
    threshold: float = 0.5,
) -> npt.NDArray[np.float_]:
    r"""Calculate the overall propulsion efficiency (OPE).

    Negative OPE values can occur during the descent phase and is clipped to a
    lower bound of 0, while an upper bound of ``threshold`` is also applied.
    The most efficient engines today do not exceed this value.

    Parameters
    ----------
    true_airspeed: npt.NDArray[np.float_]
        True airspeed for each waypoint, [:math:`m s^{-1}`].
    F_thrust: npt.NDArray[np.float_]
        Thrust force provided by the engine, [:math:`N`].
    fuel_flow: npt.NDArray[np.float_]
        Fuel mass flow rate, [:math:`kg s^{-1}`].
    q_fuel : float
        Lower calorific value (LCV) of fuel, [:math:`J \ kg_{fuel}^{-1}`].
    is_descent : npt.NDArray[np.float_] | None
        Boolean array that indicates if a waypoint is in a descent phase.
    threshold : float
        Upper bound for realistic engine efficiency.

    Returns
    -------
    npt.NDArray[np.float_]
        Overall propulsion efficiency (OPE)

    References
    ----------
    - :cite:`schumannConditionsContrailFormation1996`
    - :cite:`cumpstyJetPropulsion2015`
    """
    ope = (F_thrust * true_airspeed) / (fuel_flow * q_fuel)
    if is_descent is not None:
        ope[is_descent] = 0.0

    n_unrealistic = np.sum(ope > threshold)
    if n_unrealistic:
        logger.debug(
            "Found %s engine efficiency values exceeding %s. These are clipped.",
            n_unrealistic,
            threshold,
        )
    ope.clip(0.0, threshold, out=ope)  # clip in place
    return ope


# -------------------
# Aircraft fuel quantities
# -------------------


def fuel_burn(
    fuel_flow: npt.NDArray[np.float_], segment_duration: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """Calculate the fuel consumption at each waypoint.

    Parameters
    ----------
    fuel_flow: npt.NDArray[np.float_]
        Fuel mass flow rate, [:math:`kg s^{-1}`]
    segment_duration: npt.NDArray[np.float_]
        Time difference between waypoints, [:math:`s`]

    Returns
    -------
    npt.NDArray[np.float_]
        Fuel consumption at each waypoint, [:math:`kg`]
    """
    return fuel_flow * segment_duration


def equivalent_fuel_flow_rate_at_sea_level(
    fuel_flow_cruise: npt.NDArray[np.float_],
    theta_amb: npt.NDArray[np.float_],
    delta_amb: npt.NDArray[np.float_],
    mach_num: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    r"""Convert fuel mass flow rate at cruise conditions to equivalent flow rate at sea level.

    Refer to Eq. (40) in :cite:`duboisFuelFlowMethod22006`.

    Parameters
    ----------
    fuel_flow_cruise : npt.NDArray[np.float_]
        Fuel mass flow rate per engine, [:math:`kg s^{-1}`]
    theta_amb : npt.NDArray[np.float_]
        Ratio of the ambient temperature to the temperature at mean sea-level.
    delta_amb : npt.NDArray[np.float_]
        Ratio of the pressure altitude to the surface pressure.
    mach_num : npt.NDArray[np.float_]
        Mach number, [:math: `Ma`]

    Returns
    -------
    npt.NDArray[np.float_]
        Estimate of fuel flow per engine at sea level, [:math:`kg \ s^{-1}`].

    References
    ----------
    - :cite:`duboisFuelFlowMethod22006`
    """
    return fuel_flow_cruise * (theta_amb**3.8 / delta_amb) * np.exp(0.2 * mach_num**2)


def reserve_fuel_requirements(
    rocd: npt.NDArray[np.float_],
    altitude_ft: npt.NDArray[np.float_],
    fuel_flow: npt.NDArray[np.float_],
    fuel_burn: npt.NDArray[np.float_],
) -> float:
    r"""
    Estimate reserve fuel requirements.

    Parameters
    ----------
    rocd: npt.NDArray[np.float_]
        Rate of climb and descent, [:math:`ft \ min^{-1}`]
    altitude_ft: npt.NDArray[np.float_]
        Altitude, [:math:`ft`]
    fuel_flow: npt.NDArray[np.float_]
        Fuel mass flow rate, [:math:`kg \ s^{-1}`].
    fuel_burn: npt.NDArray[np.float_]
        Fuel consumption for each waypoint, [:math:`kg`]

    Returns
    -------
    npt.NDArray[np.float_]
        Reserve fuel requirements, [:math:`kg`]

    References
    ----------
    - :cite:`wasiukAircraftPerformanceModel2015`

    Notes
    -----
    The real-world calculation of the reserve fuel requirements is highly complex
    (refer to Section 2.3.3 of :cite:`wasiukAircraftPerformanceModel2015`).
    This implementation is simplified by taking the maximum between the following two conditions:

    1. Fuel required to fly +90 minutes at the main cruise altitude at the end of the
       cruise aircraft weight.
    2. Uplift the total fuel consumption for the flight by +15%

    See Also
    --------
    :func:`flight.segment_phase`
    :func:`fuel_burn`
    """
    segment_phase = flight.segment_phase(rocd, altitude_ft)

    # In case flight does not have cruise phase
    is_climb_cruise = (segment_phase == flight.FlightPhase.CLIMB) | (
        segment_phase == flight.FlightPhase.CRUISE
    )

    # If there are no climb and cruise phase, take the mean
    if not np.all(is_climb_cruise):
        ff_end_of_cruise = np.nanmean(fuel_flow)

    # Take average of final three waypoints
    else:
        ff_end_of_cruise = np.nanmean(fuel_flow[is_climb_cruise][-3:])

    reserve_fuel_1 = ff_end_of_cruise * (90 * 60)
    reserve_fuel_2 = 0.15 * float(np.nansum(fuel_burn))
    return np.maximum(reserve_fuel_1, reserve_fuel_2)


# -------------------
# Aircraft mass
# -------------------


def aircraft_weight(aircraft_mass: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    """Calculate the aircraft weight at each waypoint.

    Parameters
    ----------
    aircraft_mass : npt.NDArray[np.float_]
        Aircraft mass, [:math:`kg`]

    Returns
    -------
    npt.NDArray[np.float_]
        Aircraft weight, [:math:`N`]
    """
    return aircraft_mass * constants.g


def initial_aircraft_mass(
    *,
    operating_empty_weight: float,
    max_takeoff_weight: float,
    max_payload: float,
    total_fuel_burn: float,
    total_reserve_fuel: float,
    load_factor: float,
) -> float:
    """Estimate initial aircraft mass as a function of load factor and fuel requirements.

    Parameters
    ----------
    operating_empty_weight: float
        Aircraft operating empty weight, i.e. the basic weight of an aircraft including
        the crew and necessary equipment, but excluding usable fuel and payload, [:math:`kg`]
    max_takeoff_weight: float
        Aircraft maximum take-off weight, [:math:`kg`]
    max_payload: float
        Aircraft maximum payload, [:math:`kg`]
    total_fuel_burn: float
        Total fuel consumption for the flight, obtained from prior iterations, [:math:`kg`]
    total_reserve_fuel: float
        Total reserve fuel requirements, [:math:`kg`]
    load_factor: float
        Aircraft load factor assumption (between 0 and 1)

    Returns
    -------
    float
        Aircraft mass at the initial waypoint, [:math:`kg`]

    References
    ----------
    - :cite:`wasiukAircraftPerformanceModel2015`

    See Also
    --------
    :func:`reserve_fuel_requirements`
    """
    initial_amass = (
        operating_empty_weight
        + (load_factor * max_payload)
        + (total_fuel_burn + total_reserve_fuel)
    )
    return np.minimum(initial_amass, max_takeoff_weight)


def update_aircraft_mass(
    *,
    operating_empty_weight: float,
    ref_mass: float,
    max_takeoff_weight: float,
    max_payload: float,
    fuel_burn: npt.NDArray[np.float_],
    total_reserve_fuel: float,
    load_factor: None | float = None,
) -> npt.NDArray[np.float_]:
    """Update aircraft mass based on the simulated total fuel consumption.

    Used internally for finding aircraft mass iteratively.

    Parameters
    ----------
    operating_empty_weight: float
        Aircraft operating empty weight, i.e. the basic weight of an aircraft including
        the crew and necessary equipment, but excluding usable fuel and payload, [:math:`kg`].
    ref_mass: float
        Aircraft reference mass, [:math:`kg`].
    max_takeoff_weight: float
        Aircraft maximum take-off weight, [:math:`kg`].
    max_payload: float
        Aircraft maximum payload, [:math:`kg`]
    fuel_burn: npt.NDArray[np.float_]
        Fuel consumption for each waypoint, [:math:`kg`]
    total_reserve_fuel: float
        Total reserve fuel requirements, [:math:`kg`]
    load_factor: None | float
        Aircraft load factor assumption (between 0 and 1). If None is given, the reference mass
        from the BADA 3 database will be used to get the mass of the specific aircraft type.

    Returns
    -------
    npt.NDArray[np.float_]
        Updated aircraft mass, [:math:`kg`]

    See Also
    --------
    :func:`fuel_burn`
    :func:`reserve_fuel_requirements`
    """
    if load_factor is not None:
        initial_amass = initial_aircraft_mass(
            operating_empty_weight=operating_empty_weight,
            max_takeoff_weight=max_takeoff_weight,
            max_payload=max_payload,
            total_fuel_burn=float(np.nansum(fuel_burn)),
            total_reserve_fuel=total_reserve_fuel,
            load_factor=load_factor,
        )
    else:
        initial_amass = ref_mass

    # Calculate updated aircraft mass for each waypoint
    amass = np.empty_like(fuel_burn)
    amass[0] = initial_amass
    amass[1:] = initial_amass - np.nancumsum(fuel_burn)[:-1]
    return amass


# ------------------------------------------------------------------
# Temperature and pressure at different sections of the jet engine
# ------------------------------------------------------------------


def compressor_inlet_temperature(T: ArrayScalarLike, mach_num: ArrayScalarLike) -> ArrayScalarLike:
    """Calculate compressor inlet temperature for Jet engine, :math:`T_{2}`.

    Parameters
    ----------
    T : ArrayScalarLike
        Ambient temperature, [:math:`K`]
    mach_num : ArrayScalarLike
        Mach number

    Returns
    -------
    ArrayScalarLike
        Compressor inlet temperature, [:math:`K`]

    References
    ----------
    - :cite:`stettlerGlobalCivilAviation2013`
    - :cite:`cumpstyJetPropulsion2015`
    """
    return T * (1 + ((constants.kappa - 1) / 2) * mach_num**2)


def compressor_inlet_pressure(p: ArrayScalarLike, mach_num: ArrayScalarLike) -> ArrayScalarLike:
    """Calculate compressor inlet pressure for Jet engine, :math:`P_{2}`.

    Parameters
    ----------
    p : ArrayScalarLike
        Ambient pressure, [:math:`Pa`]
    mach_num : ArrayScalarLike
        Mach number

    Returns
    -------
    ArrayScalarLike
        Compressor inlet pressure, [:math:`Pa`]

    References
    ----------
    - :cite:`stettlerGlobalCivilAviation2013`
    - :cite:`cumpstyJetPropulsion2015`
    """
    power_term = constants.kappa / (constants.kappa - 1)
    return p * (1 + ((constants.kappa - 1) / 2) * mach_num**2) ** power_term


def combustor_inlet_pressure(
    pressure_ratio: float,
    p_comp_inlet: ArrayScalarLike,
    thrust_setting: ArrayScalarLike,
) -> ArrayScalarLike:
    """Calculate combustor inlet pressure, :math:`P_{3}`.

    Parameters
    ----------
    pressure_ratio : float
        Engine pressure ratio, unitless
    p_comp_inlet : ArrayScalarLike
        Compressor inlet pressure, [:math:`Pa`]
    thrust_setting : ArrayScalarLike
        Engine thrust setting, unitless

    Returns
    -------
    ArrayScalarLike
        Combustor inlet pressure, [:math:`Pa`]

    References
    ----------
    - :cite:`stettlerGlobalCivilAviation2013`
    - :cite:`cumpstyJetPropulsion2015`
    """
    return (p_comp_inlet * (pressure_ratio - 1) * thrust_setting) + p_comp_inlet


def combustor_inlet_temperature(
    comp_efficiency: float,
    T_comp_inlet: ArrayScalarLike,
    p_comp_inlet: ArrayScalarLike,
    p_comb_inlet: ArrayScalarLike,
) -> ArrayScalarLike:
    """Calculate combustor inlet temperature, :math:`T_{3}`.

    Parameters
    ----------
    comp_efficiency : float
        Engine compressor efficiency, [:math:`0 - 1`]
    T_comp_inlet : ArrayScalarLike
        Compressor inlet temperature, [:math:`K`]
    p_comp_inlet : ArrayScalarLike
        Compressor inlet pressure, [:math:`Pa`]
    p_comb_inlet : ArrayScalarLike
        Compressor inlet pressure, [:math:`Pa`]

    Returns
    -------
    ArrayScalarLike
        Combustor inlet temperature, [:math:`K`]

    References
    ----------
    - :cite:`stettlerGlobalCivilAviation2013`
    - :cite:`cumpstyJetPropulsion2015`
    """
    power_term = (constants.kappa - 1) / (constants.kappa * comp_efficiency)
    return T_comp_inlet * (p_comb_inlet / p_comp_inlet) ** power_term


def turbine_inlet_temperature(
    afr: ArrayScalarLike, T_comb_inlet: ArrayScalarLike, q_fuel: float
) -> ArrayScalarLike:
    r"""Calculate turbine inlet temperature, :math:`T_{4}`.

    Parameters
    ----------
    afr : ArrayScalarLike
        Air-to-fuel ratio, unitless
    T_comb_inlet : ArrayScalarLike
        Combustor inlet temperature, [:math:`K`]
    q_fuel : float
        Lower calorific value (LCV) of fuel, :math:`[J \ kg_{fuel}^{-1}]`

    Returns
    -------
    ArrayScalarLike
        Tubrine inlet temperature, [:math:`K`]

    References
    ----------
    - :cite:`cumpstyJetPropulsion2015`
    """
    return (afr * constants.c_pd * T_comb_inlet + q_fuel) / (constants.c_p_combustion * (1 + afr))


# --------------------------------------
# Engine thrust force and thrust settings
# --------------------------------------


def thrust_force(
    altitude: npt.NDArray[np.float_],
    true_airspeed: npt.NDArray[np.float_],
    dt: npt.NDArray[np.float_],
    aircraft_mass: npt.NDArray[np.float_],
    F_drag: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    r"""Calculate the thrust force at each waypoint.

    Parameters
    ----------
    altitude : npt.NDArray[np.float_]
        Waypoint altitude, [:math:`m`]
    true_airspeed : npt.NDArray[np.float_]
        True airspeed, [:math:`m \ s^{-1}`]
    dt : npt.NDArray[np.float_]
        Time between waypoints, [:math:`s`]
    aircraft_mass : npt.NDArray[np.float_]
        Aircraft mass, [:math:`kg`]
    F_drag : npt.NDArray[np.float_]
        Draft force, [:math:`N`]

    Returns
    -------
    npt.NDArray[np.float_]
        Thrust force, [:math:`N`]

    References
    ----------
    - :cite:`eurocontrolUSERMANUALBASE2010`

    Notes
    -----
    The model balances the rate of forces acting on the aircraft
    with the rate in increase in energy.

    This estimate of thrust force is used in the BADA Total-Energy Model (Eq. 3.2-1).

    Negative thrust must be corrected.
    """
    dh_dt = np.empty_like(altitude)
    dh_dt[:-1] = np.diff(altitude) / dt[:-1]
    dh_dt[-1] = 0.0
    np.nan_to_num(dh_dt, copy=False)

    dv_dt = np.empty_like(true_airspeed)
    dv_dt[:-1] = np.diff(true_airspeed) / dt[:-1]
    dv_dt[-1] = 0.0
    np.nan_to_num(dv_dt, copy=False)

    return (
        F_drag
        + (aircraft_mass * constants.g * dh_dt + aircraft_mass * true_airspeed * dv_dt)
        / true_airspeed
    )


def thrust_setting_nd(
    true_airspeed: ArrayScalarLike,
    thrust_setting: ArrayScalarLike,
    T: ArrayScalarLike,
    p: ArrayScalarLike,
    pressure_ratio: float,
    q_fuel: float,
    *,
    comp_efficiency: float = 0.9,
    cruise: bool = False,
) -> ArrayScalarLike:
    r"""Calculate the non-dimensionalized thrust setting of a Jet engine.

    Result is in terms of the ratio of turbine inlet to the
    compressor inlet temperature (t4_t2)

    Parameters
    ----------
    true_airspeed : ArrayScalarLike
        True airspeed, [:math:`m \ s^{-1}`]
    thrust_setting : ArrayScalarLike
        Engine thrust setting, unitless
    T : ArrayScalarLike
        Ambient temperature, [:math:`K`]
    p : ArrayScalarLike
        Ambient pressure, [:math:`Pa`]
    pressure_ratio : float
        Engine pressure ratio, unitless
    q_fuel : float
        Lower calorific value (LCV) of fuel, :math:`[J \ kg_{fuel}^{-1}]`
    comp_efficiency : float, optional
        Engine compressor efficiency, [:math:`0 - 1`].
        Defaults to 0.9
    cruise : bool, optional
        Defaults to False

    Returns
    -------
    ArrayScalarLike
        Ratio of turbine inlet to the compressor inlet temperature, unitless

    References
    ----------
    - :cite:`cumpstyJetPropulsion2015`
    - :cite:`teohAviationContrailClimate2022`
    """
    mach_num = units.tas_to_mach_number(true_airspeed, T)
    T_compressor_inlet = compressor_inlet_temperature(T, mach_num)
    p_compressor_inlet = compressor_inlet_pressure(p, mach_num)
    p_combustor_inlet = combustor_inlet_pressure(pressure_ratio, p_compressor_inlet, thrust_setting)
    T_combustor_inlet = combustor_inlet_temperature(
        comp_efficiency, T_compressor_inlet, p_compressor_inlet, p_combustor_inlet
    )
    afr = air_to_fuel_ratio(thrust_setting, cruise=cruise, T_compressor_inlet=T_compressor_inlet)
    T_turbine_inlet = turbine_inlet_temperature(afr, T_combustor_inlet, q_fuel)
    return T_turbine_inlet / T_compressor_inlet


def air_to_fuel_ratio(
    thrust_setting: ArrayScalarLike,
    *,
    cruise: bool = False,
    T_compressor_inlet: None | ArrayScalarLike = None,
) -> ArrayScalarLike:
    """Calculate air-to-fuel ratio from thrust setting.

    Parameters
    ----------
    thrust_setting : ArrayScalarLike
        Engine thrust setting, unitless
    cruise : bool
        Estimate thrust setting for cruise conditions. Defaults to False.
    T_compressor_inlet : None | ArrayScalarLike
        Compressor inlet temperature, [:math:`K`]
        Required if ``cruise`` is True.
        Defaults to None

    Returns
    -------
    ArrayScalarLike
        Air-to-fuel ratio, unitless

    References
    ----------
    - :cite:`cumpstyJetPropulsion2015`
    - AFR equation from :cite:`stettlerGlobalCivilAviation2013`
    - Scaling factor to cruise from Eq. (30) of :cite:`duboisFuelFlowMethod22006`

    """
    afr = (0.0121 * thrust_setting + 0.008) ** (-1)

    if not cruise:
        return afr

    if T_compressor_inlet is None:
        raise ValueError("`T_compressor_inlet` is required when `cruise` is True")

    return afr * (T_compressor_inlet / constants.T_msl)


# -------------------
# Atmospheric ratios
# -------------------


def temperature_ratio(T: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    """Calculate the ratio of ambient temperature relative to the temperature at mean sea level.

    Parameters
    ----------
    T : npt.NDArray[np.float_]
        Air temperature, [:math:`K`]

    Returns
    -------
    npt.NDArray[np.float_]
        Ratio of the temperature to the temperature at mean sea-level (MSL).
    """
    return T / constants.T_msl


def pressure_ratio(p: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    """Calculate the ratio of ambient pressure relative to the surface pressure.

    Parameters
    ----------
    p : npt.NDArray[np.float_]
        Air pressure, [:math:`Pa`]

    Returns
    -------
    npt.NDArray[np.float_]
        Ratio of the pressure altitude to the surface pressure.
    """
    return p / constants.p_surface


def density_ratio(rho: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    r"""Calculate the ratio of air density relative to the air density at mean-sea-level.

    Parameters
    ----------
    rho : npt.NDArray[np.float_]
        Air density, [:math:`kg \ m^{3}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Ratio of the density to the air density at mean sea-level (MSL).
    """
    return rho / constants.rho_msl



"""Test Fleet data structure."""

import numpy as np
import pytest

from pycontrails import Flight
from pycontrails.core.fleet import Fleet

try:
    from pycontrails.utils.synthetic_flight import SyntheticFlight
except ImportError:
    pytest.skip("SyntheticFlight not available", allow_module_level=True)


@pytest.fixture(scope="module")
def syn():
    """Build `SyntheticFlight` generator."""
    bounds = {
        "time": (np.datetime64("2019-10-05"), np.datetime64("2019-10-05T12")),
        "longitude": (-20, 20),
        "latitude": (30, 60),
        "level": (200, 300),
    }
    return SyntheticFlight(bounds, speed_m_per_s=220, aircraft_type="A320", seed=19)


def test_fleet_io(syn: SyntheticFlight):
    """Confirm that a fleet can be instantiated from a list of `Flights`.

    Check that the flights coming out of method `to_flight_list` agree with inputs.
    """
    in_fls = [syn() for _ in range(100)]
    fleet = Fleet.from_seq(in_fls)
    assert fleet.n_flights == 100

    out_fls = fleet.to_flight_list()
    for fl1, fl2 in zip(in_fls, out_fls):
        assert fl1 == fl2


def test_fleet_calc_final_waypoints(syn: SyntheticFlight):
    """Confirm `calc_final_waypoints` method."""
    fls = [syn() for _ in range(25)]
    fleet = Fleet.from_seq(fls)
    assert isinstance(fleet.final_waypoints, np.ndarray)
    assert np.sum(fleet.final_waypoints) == len(fls)
    assert fleet.final_waypoints.dtype == bool
    idx = np.cumsum([fl.size for fl in fls]) - 1
    assert np.all(fleet.final_waypoints[idx])

    fls.append(fls[3])
    with pytest.raises(ValueError, match="Duplicate `flight_id`"):
        Fleet.from_seq(fls)
    data = fleet.data
    data["flight_id"][3] = data["flight_id"][-3]
    match = "Fleet must have contiguous waypoints blocks with constant"
    with pytest.raises(ValueError, match=match):
        Fleet(data, attrs=fleet.attrs)

    with pytest.raises(ValueError, match="Require key 'fl_attrs'"):
        Fleet(data)


def test_fleet_segment_length(syn: SyntheticFlight):
    """Check compatibility of `segment_length` and `segment_angle` on `Fleet`."""
    fls = [syn() for _ in range(20)]
    fleet = Fleet.from_seq(fls)
    assert fleet.n_flights == 20

    sl1 = fleet.segment_length()
    sl2 = np.concatenate([fl.segment_length() for fl in fls])
    np.testing.assert_array_equal(sl1, sl2)

    _, cos1 = fleet.segment_angle()
    cos2 = np.concatenate([fl.segment_angle()[1] for fl in fls])
    np.testing.assert_array_equal(cos1, cos2)


def test_fleet_run_all_methods(syn: SyntheticFlight):
    """Ensure a few common methods on `Fleet` instance run without error."""
    fls = [syn() for _ in range(30)]
    fleet = Fleet.from_seq(fls)
    assert fleet.n_flights == 30

    assert isinstance(fleet.coords, dict)
    assert isinstance(fleet.segment_angle(), tuple)
    assert isinstance(fleet.altitude_ft, np.ndarray)
    assert isinstance(fleet.level, np.ndarray)


def test_fleet_init_from_gen(syn: SyntheticFlight):
    """Check that Fleet can be instantiated from a generator of `Flights`."""
    fls = (syn() for _ in range(200))
    fleet = Fleet.from_seq(fls)
    assert fleet.n_flights == 200


def test_fleet_waypoints(syn: SyntheticFlight):
    """Check Cocip continuity conventions on `Fleet` instance."""
    fls = (syn() for _ in range(321))
    fleet = Fleet.from_seq(fls)
    assert fleet.n_flights == 321
    assert "waypoint" in fleet
    assert fleet["waypoint"].dtype == int

    # copied from `Cocip._calc_continuous`
    continuous = np.append((np.diff(fleet["waypoint"]) == 1).astype(bool), False)
    assert np.sum(~continuous) == 321  # each flight-change generates a discontinuity

    flight_ids = fleet.filter(~continuous)["flight_id"]
    # expect one of each flight_id
    for id1, id2 in zip(flight_ids, fleet.attrs["fl_attrs"]):
        assert id1 == id2


def test_duplicate_flight_ids(syn: SyntheticFlight):
    """Check that Fleet raises exception if duplicate flight id is found."""
    fls = [syn() for _ in range(222)]
    fls[151].attrs.update(flight_id=333)
    fls[97].attrs.update(flight_id=333)
    with pytest.raises(ValueError, match="Duplicate `flight_id` 333 found."):
        Fleet.from_seq(fls)


def test_type_of_fleet(syn: SyntheticFlight):
    """Confirm that a `Fleet` is an instance of `Flight`, but an iterable of `Flight` is not.

    This logic is used in `Cocip.eval` method.
    """
    fls = [syn() for _ in range(100)]
    fleet = Fleet.from_seq(fls)
    fl = fls[0]

    # This logic can be used to distinguish Iterable[Flight] from Flight
    assert isinstance(fl, Flight)
    assert isinstance(fleet, Flight)
    assert not isinstance(fls, Flight)

    # NOTE: All have iter method, so this cannot be used to distinguish
    assert hasattr(fl, "__iter__")
    assert hasattr(fleet, "__iter__")
    assert hasattr(fls, "__iter__")


def test_fleet_numeric_types_to_data(syn: SyntheticFlight):
    """Check that numeric attributes are attached to underlying `Fleet` data."""
    fls = [syn() for _ in range(40)]
    for fl in fls:
        fl.attrs["numeric1"] = np.random.random()  # not even seeding it
        fl.attrs["numeric2"] = np.random.randint(5)
        fl.attrs["non-numeric"] = "contrail"
    fleet = Fleet.from_seq(fls)
    fleet.ensure_vars(
        [
            "waypoint",
            "longitude",
            "latitude",
            "level",
            "time",
            "flight_id",
            "numeric1",
            "numeric2",
        ]
    )
    assert "non-numeric" not in fleet

    # Nothing in fleet.attrs is numeric
    n_keys = len(fleet.data.keys())
    fleet.broadcast_numeric_attrs()
    assert len(fleet.data.keys()) == n_keys


def test_fleet_true_airspeed(syn: SyntheticFlight):
    """Confirm that method segment_true_airspeed on Fleet agrees with Flight-wise call."""
    fls = [syn() for _ in range(44)]
    fleet = Fleet.from_seq(fls)
    rng = np.random.default_rng(1234)
    u_wind = rng.uniform(-5, 5, fleet.size)
    v_wind = rng.uniform(-5, 5, fleet.size)
    with pytest.raises(NotImplementedError):
        fleet.segment_groundspeed(u_wind, v_wind)

    tas = fleet.segment_true_airspeed(u_wind, v_wind)
    assert tas.size == fleet.size

    # NaN values occur just before waypoints reset
    expected_nan = np.diff(fleet["waypoint"], append=np.nan) != 1
    np.testing.assert_array_equal(expected_nan, np.isnan(tas))

    i0 = 0
    for fl in fls:
        i1 = i0 + fl.size
        slice_ = slice(i0, i1)
        u_wind_slice = u_wind[slice_]
        v_wind_slice = v_wind[slice_]
        tas_slice = fl.segment_true_airspeed(u_wind_slice, v_wind_slice)
        np.testing.assert_array_equal(tas_slice, tas[slice_])
        i0 = i1



"""Default parameters for CoCiP models.

Used by :class:`Cocip` and :class:`CocipGrid`.
"""

from __future__ import annotations

import dataclasses

import numpy as np

from pycontrails.core.models import ModelParams
from pycontrails.models.aircraft_performance import AircraftPerformance
from pycontrails.models.emissions.emissions import EmissionsParams
from pycontrails.models.humidity_scaling import HumidityScaling


@dataclasses.dataclass
class CocipParams(ModelParams):
    """Model parameters required by the CoCiP models."""

    # -------------------------
    # Implementation parameters
    # -------------------------

    #: Determines whether :meth:`Cocip.process_emissions` runs on model :meth:`Cocip.eval`
    #: Set to ``False`` when input Flight includes emissions data.
    process_emissions: bool = True

    #: Integrate initial_contrails with time steps of dt
    dt_integration: np.timedelta64 = np.timedelta64(30, "m")

    #: Difference in altitude between top and bottom layer for stratification calculations (m)
    #: Used to approximate derivative of "lagrangian_tendency_of_air_pressure" layer
    dz_m: float = 200.0

    #: Vertical resolution (m) associated to met data.
    #: Constant below applies to ECMWF data.
    effective_vertical_resolution: float = 2000.0

    #: Shift the time coordinates of radiation parameters for accumulated values
    #: TODO: change this to np.timedelta64(0, "m") when we start using other datasets
    shift_radiation_time: np.timedelta64 = -np.timedelta64(30, "m")

    #: Smoothing parameters for true airspeed.
    #: Only used for Flight models.
    #: Passed directly to :func:`scipy.signal.savgol_filter`.
    #: See :meth:`pycontrails.Flight.segment_true_airspeed` for details.
    smooth_true_airspeed: bool = True
    smooth_true_airspeed_window_length: int = 7
    smooth_true_airspeed_polyorder: int = 1

    #: Humidity scaling
    humidity_scaling: HumidityScaling | None = None

    # --------------
    # Downselect met
    # --------------

    #: Met longitude [WGS84] buffer for Cocip evolution.
    met_longitude_buffer: tuple[float, float] = (10.0, 10.0)

    #: Met latitude buffer [WGS84] for Cocip evolution.
    met_latitude_buffer: tuple[float, float] = (10.0, 10.0)

    #: Met level buffer [:math:`hPa`] for Cocip initialization and evolution.
    met_level_buffer: tuple[float, float] = (200.0, 200.0)

    # ---------
    # Filtering
    # ---------

    #: Filter out waypoints if the don't satisfy the SAC criteria
    #: Note that the SAC algorithm will still be run to calculate
    #: ``T_critical_sac`` for use estimating initial ice particle number.
    #: Passing in a non-default value is unusual, but is included
    #: to allow for false negative calibration and model uncertainty studies.
    filter_sac: bool = True

    #: Filter out waypoints if they don't satisfy the initial persistent criteria
    #: Passing in a non-default value is unusual, but is included
    #: to allow for false negative calibration and model uncertainty studies.
    filter_initially_persistent: bool = True

    #: Continue evolving contrail waypoints ``persistent_buffer`` beyond
    #: end of contrail life.
    #: Passing in a non-default value is unusual, but is included
    #: to allow for false negative calibration and model uncertainty studies.
    persistent_buffer: np.timedelta64 | None = None

    # -------
    # Outputs
    # -------

    #: Add additional values to the flight and contrail that are not explicitly
    #: necessary for calculation.
    #: See also :attr:`CocipGridParams.verbose_outputs_formation` and
    #: :attr:`CocipGridParams.verbose_outputs_evolution`.
    verbose_outputs: bool = False

    # ----------------
    # Model parameters
    # ----------------

    #: Initial wake vortex depth scaling factor.
    #: This factor scales max contrail downward displacement after the wake vortex phase
    #: to set the initial contrail depth.
    #: Denoted :math:`C_{D0}` in eq (14) in :cite:`schumannContrailCirrusPrediction2012`.
    initial_wake_vortex_depth: float = 0.5

    #: Sedimentation impact factor. Denoted by :math:`f_{T}` in eq. (35) of
    #: :cite:`schumannContrailCirrusPrediction2012`.
    #: Schumann describes this as "an important adjustable parameter", and sets
    #: it to 0.1 in the original publication.
    sedimentation_impact_factor: float = 0.5

    #: Default ``nvpm_ei_n`` value if no data provided and emissions calculations fails.
    default_nvpm_ei_n: float = EmissionsParams.default_nvpm_ei_n

    #: Parameter denoted by :math:`n` in eq. (39) of :cite:`schumannContrailCirrusPrediction2012`.
    wind_shear_enhancement_exponent: float = 0.5

    #: Multiply flight black carbon number by enhancement factor.
    #: A value of 1.0 provides no scaling.
    #: Primarily used to support uncertainty estimation.
    nvpm_ei_n_enhancement_factor: float = 1.0

    #: Lower bound for ``nvpm_ei_n`` to account for ambient aerosol
    #: particles for newer engines, [:math:`kg^{-1}`]
    min_ice_particle_number_nvpm_ei_n: float = 1e13

    #: Upper bound for contrail plume depth, constraining it to realistic values
    #: CoCiP only uses the ambient conditions at the mid-point of the Gaussian plume,
    #: and the edges could be in subsaturated conditions and sublimate. Important when
    #: :attr:`radiative_heating_effects` is enabled.
    max_contrail_depth: float = 1500.0

    #: Experimental. Radiative heating effects on contrail cirrus properties.
    #: Terrestrial and solar radiances warm the contrail ice particles and cause
    #: convective turbulence. This effect is expected to enhance vertical mixing
    #: and reduce the lifetime of contrail cirrus. This parameter is experimental,
    #: and the CoCiP implementation of this parameter may change.
    #:
    #:  .. versionadded:: 0.28.9
    radiative_heating_effects: bool = False

    #: Radius threshold for regime bins, [:math:`\mu m`]
    #: This is the row index label for ``habit_distributions``.
    #: See Table 2 in :cite:`schumannEffectiveRadiusIce2011`.
    radius_threshold_um: np.ndarray = dataclasses.field(
        default_factory=lambda: np.array([5.0, 9.5, 23.0, 190.0, 310.0], dtype=np.float32)
    )

    #: Particle habit (shape) types.
    #: This is the column index label for ``habit_distributions``.
    #: See Table 2 in :cite:`schumannEffectiveRadiusIce2011`.
    habits: np.ndarray = dataclasses.field(
        default_factory=lambda: np.array(
            [
                "Sphere",
                "Solid column",
                "Hollow column",
                "Rough aggregate",
                "Rosette-6",
                "Plate",
                "Droxtal",
                "Myhre",
            ]
        )
    )

    #: Mix of ice particle habits in each radius regime.
    #: Rows indexes are ``radius_threshold_um`` elements.
    #: Columns indexes are ``habits`` particle habit type.
    #: See Table 2 from :cite:`schumannEffectiveRadiusIce2011`.
    habit_distributions: np.ndarray = dataclasses.field(
        default_factory=lambda: np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                [0.0, 0.3, 0.0, 0.0, 0.0, 0.0, 0.7, 0.0],
                [0.0, 0.3, 0.0, 0.0, 0.3, 0.0, 0.4, 0.0],
                [0.0, 0.5, 0.0, 0.0, 0.15, 0.35, 0.0, 0.0],
                [0.0, 0.45, 0.45, 0.1, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.03, 0.97, 0.0, 0.0, 0.0],
            ],
            dtype=np.float32,
        )
    )

    #: Scale shortwave radiative forcing.
    #: Primarily used to support uncertainty estimation.
    rf_sw_enhancement_factor: float = 1.0

    #: Scale longwave radiative forcing.
    #: Primarily used to support uncertainty estimation.
    rf_lw_enhancement_factor: float = 1.0

    # ---------------------------------------
    # Conditions for end of contrail lifetime
    # ---------------------------------------

    #: Minimum altitude domain in simulation, [:math:`m`]
    #: If set to ``None``, this check is disabled.
    min_altitude_m: float | None = 6000.0

    #: Maximum altitude domain in simulation, [:math:`m`]
    #: If set to ``None``, this check is disabled.
    max_altitude_m: float | None = 13000.0

    #: Maximum contrail segment length in simulation to prevent unrealistic values, [:math:`m`].
    max_seg_length_m: float = 40000.0

    #: Max age of contrail evolution.
    max_age: np.timedelta64 = np.timedelta64(20, "h")

    #: Minimum contrail optical depth.
    min_tau: float = 1e-6

    #: Maximum contrail optical depth to prevent unrealistic values.
    max_tau: float = 1e10

    #: Minimum contrail ice particle number per volume of air.
    min_n_ice_per_m3: float = 1e3

    #: Maximum contrail ice particle number per volume of air to prevent unrealistic values.
    max_n_ice_per_m3: float = 1e20


@dataclasses.dataclass
class CocipFlightParams(CocipParams):
    """Flight specific CoCiP parameters."""

    #: Aircraft performance model
    aircraft_performance: AircraftPerformance | None = None



"""Test the fuel module."""

import dataclasses

import pytest

from pycontrails import Fuel, HydrogenFuel, JetA, SAFBlend


def test_fuel_and_saf_properties():
    """Check basic properties of JetA and SAFBlend."""
    jet_a = JetA()
    saf = SAFBlend(10)
    assert isinstance(jet_a, Fuel)
    assert isinstance(saf, Fuel)

    # Every attribute on jet_a is also on saf
    assert jet_a.__dict__.keys() - saf.__dict__.keys() == set()

    # And the saf instance has one more attribute
    saf.__dict__.keys() - jet_a.__dict__.keys() == {"pct_blend"}

    assert saf.fuel_name == "Jet A-1 / Sustainable Aviation Fuel Blend"
    assert jet_a.fuel_name == "Jet A-1"


def test_saf_zero_pct_blend():
    """Confirm that SAFBlend(pct_blend=0) is identical to JetA."""
    jet_a = JetA()
    saf_0 = SAFBlend(pct_blend=0)

    for field in dataclasses.fields(jet_a):
        k = field.name
        if k == "fuel_name":
            continue
        assert getattr(jet_a, k) == getattr(saf_0, k)


def test_saf_blend_values():
    """Confirm a few pinned values for `SAFBlend` class."""
    saf_50 = SAFBlend(pct_blend=50)
    saf_100 = SAFBlend(pct_blend=100)

    assert isinstance(saf_50, Fuel)
    assert saf_50.hydrogen_content == 14.55
    assert saf_50.q_fuel == 43.665e6
    assert saf_50.ei_h2o == pytest.approx(1.296848, abs=1e-3)

    assert isinstance(saf_100, Fuel)
    assert saf_100.hydrogen_content == 15.3
    assert saf_100.q_fuel == 44.2e6
    assert saf_100.ei_h2o == pytest.approx(1.363696, abs=1e-3)


def test_hydrogen_fuel():
    """Check some properties of HydrogenFuel."""
    hf = HydrogenFuel()
    assert hf.fuel_name == "Hydrogen"

    jet_a = JetA()
    assert isinstance(hf, Fuel)
    assert hf.q_fuel > jet_a.q_fuel
    assert hf.ei_h2o > jet_a.ei_h2o



"""Calculate nitrogen oxide (NOx), carbon monoxide (CO) and hydrocarbon (HC) emissions.

This modules applies the Fuel Flow Method 2 (FFM2) from DuBois & Paynter (2006) for a given
aircraft-engine pair.

References
----------
- :cite:`duboisFuelFlowMethod22006`
"""
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from pycontrails.core.interpolation import EmissionsProfileInterpolator
from pycontrails.physics import jet, units


def nitrogen_oxide_emissions_index_profile(
    ff_idle: float,
    ff_approach: float,
    ff_climb: float,
    ff_take_off: float,
    ei_nox_idle: float,
    ei_nox_approach: float,
    ei_nox_climb: float,
    ei_nox_take_off: float,
) -> EmissionsProfileInterpolator:
    """
    Create the nitrogen oxide (NOx) emissions index (EI) profile for the given engine type.

    Parameters
    ----------
    ff_idle: float
        ICAO EDB fuel mass flow rate at idle conditions (7% power), [:math:`kg s^{-1}`]
    ff_approach: float
        ICAO EDB fuel mass flow rate at approach (30% power), [:math:`kg s^{-1}`]
    ff_climb: float
        ICAO EDB fuel mass flow rate at climb out (85% power), [:math:`kg s^{-1}`]
    ff_take_off: float
        ICAO EDB fuel mass flow rate at take-off (100% power), [:math:`kg s^{-1}`]
    ei_nox_idle: float
        ICAO EDB NOx emissions index at idle conditions (7% power), [:math:`g_{NO_{X}}/kg_{fuel}`]
    ei_nox_approach: float
        ICAO EDB NOx emissions index at approach (30% power), [:math:`g_{NO_{X}}/kg_{fuel}`]
    ei_nox_climb: float
        ICAO EDB NOx emissions index at climb out (85% power), [:math:`g_{NO_{X}}/kg_{fuel}`]
    ei_nox_take_off: float
        ICAO EDB NOx emissions index at take-off (100% power), [:math:`g_{NO_{X}}/kg_{fuel}`]

    Returns
    -------
    EmissionsProfileInterpolator
        log of NOx emissions index versus the log of fuel mass flow rate for a given engine type

    Raises
    ------
    ValueError
        If any EI nox values are non-positive.
    """
    fuel_flow = np.array([ff_idle, ff_approach, ff_climb, ff_take_off], dtype=float)
    installation_correction_factor = np.array([1.100, 1.020, 1.013, 1.010])
    fuel_flow *= installation_correction_factor

    ei_nox = np.array([ei_nox_idle, ei_nox_approach, ei_nox_climb, ei_nox_take_off], dtype=float)

    if np.any(ei_nox <= 0):
        raise ValueError("Zero value(s) encountered in the EI NOx.")

    return EmissionsProfileInterpolator(xp=np.log(fuel_flow), fp=np.log(ei_nox))


def co_hc_emissions_index_profile(
    ff_idle: float,
    ff_approach: float,
    ff_climb: float,
    ff_take_off: float,
    ei_idle: float,
    ei_approach: float,
    ei_climb: float,
    ei_take_off: float,
) -> EmissionsProfileInterpolator:
    """Create carbon monoxide (CO) and hydrocarbon (HC) emissions index (EI) profile.

    Parameters
    ----------
    ff_idle: float
        ICAO EDB fuel mass flow rate at idle conditions
        (7% power), [:math:`kg s^{-1}`]
    ff_approach: float
        ICAO EDB fuel mass flow rate at approach
        (30% power), [:math:`kg s^{-1}`]
    ff_climb: float
        ICAO EDB fuel mass flow rate at climb out
        (85% power), [:math:`kg s^{-1}`]
    ff_take_off: float
        ICAO EDB fuel mass flow rate at take-off
        (100% power), [:math:`kg s^{-1}`]
    ei_idle: float
        ICAO EDB CO or HC emissions index at idle conditions
        (7% power), [:math:`g_{pollutant}/kg_{fuel}`]
    ei_approach: float
        ICAO EDB CO or HC emissions index at approach
        (30% power), [:math:`g_{pollutant}/kg_{fuel}`]
    ei_climb: float
        ICAO EDB CO or HC emissions index at climb out
        (85% power), [:math:`g_{pollutant}/kg_{fuel}`]
    ei_take_off: float
        ICAO EDB CO or HC emissions index at take-off
        (100% power), [:math:`g_{pollutant}/kg_{fuel}`]

    Returns
    -------
    EmissionsProfileInterpolator
        log of CO or HC emissions index versus the log of fuel mass
        flow rate for a given engine type
    """
    fuel_flow_edb = np.array([ff_idle, ff_approach, ff_climb, ff_take_off], dtype=float)
    installation_correction_factor = np.array([1.100, 1.020, 1.013, 1.010])
    fuel_flow_edb *= installation_correction_factor

    ei_edb = np.array([ei_idle, ei_approach, ei_climb, ei_take_off], dtype=float)
    min_vals_ = np.array([1e-3, 1e-3, 1e-4, 1e-4])
    ei_edb = np.maximum(ei_edb, min_vals_)

    # Get straight-line equation between idle and approach
    m, c = np.polyfit(fuel_flow_edb[:2], ei_edb[:2], deg=1)
    ei_climb_extrapolate = m * fuel_flow_edb[2] + c
    ei_hi = np.mean(ei_edb[2:])

    ff_low_power = fuel_flow_edb[3] * 0.03
    ei_co_low_power = min((m * ff_low_power + c), (2 * ei_edb[0]))
    ei_co_low_power = max(
        ei_co_low_power, 1e-3
    )  # Prevent zero/negative values, similar to line 115

    # Permutation 1: Emissions profile when the bi-linear fit does not work
    # (Figure 14 of DuBois & Paynter, 2006)
    if ei_edb[1] < ei_edb[2]:
        ff_profile = np.insert(fuel_flow_edb, 0, ff_low_power)
        ei_profile = np.array([ei_co_low_power, ei_edb[0], ei_edb[1], ei_hi, ei_hi])

    # Permutation 2: Emissions profile using a bi-linear fit (Figure 8 of DuBois & Paynter, 2006)
    elif ei_climb_extrapolate < ei_edb[2] and (m != 0):
        ff_intersect = (ei_hi - c) / m
        # Ensure intersection is between 30% and 85% fuel mass flow rate
        ff_intersect = np.clip(ff_intersect, fuel_flow_edb[1] + 0.01, fuel_flow_edb[2] - 0.01)
        ff_profile = np.array(
            [
                ff_low_power,
                fuel_flow_edb[0],
                fuel_flow_edb[1],
                ff_intersect,
                fuel_flow_edb[2],
                fuel_flow_edb[3],
            ]
        )
        ei_profile = np.array([ei_co_low_power, ei_edb[0], ei_edb[1], ei_hi, ei_hi, ei_hi])

    # Permutation 3: Point-to-point fit (Figure 13 of DuBois & Paynter, 2006)
    else:
        ff_profile = np.insert(fuel_flow_edb, 0, ff_low_power)
        ei_profile = np.insert(ei_edb, 0, ei_co_low_power)

    return EmissionsProfileInterpolator(xp=np.log(ff_profile), fp=np.log(ei_profile))


def estimate_nox(
    log_ei_nox_profile: EmissionsProfileInterpolator,
    fuel_flow_per_engine: npt.NDArray[np.float_],
    true_airspeed: npt.NDArray[np.float_],
    air_pressure: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
    specific_humidity: None | npt.NDArray[np.float_] = None,
) -> npt.NDArray[np.float_]:
    """Estimate the nitrogen oxide (NOx) emissions index (EI) at cruise conditions.

    Parameters
    ----------
    log_ei_nox_profile: EmissionsProfileInterpolator
        emissions profile containing the log of EI NOx versus log of fuel flow.
    fuel_flow_per_engine: npt.NDArray[np.float_]
        fuel mass flow rate per engine, [:math:`kg s^{-1}`]
    true_airspeed: npt.NDArray[np.float_]
        true airspeed for each waypoint, [:math:`m s^{-1}`]
    air_pressure : npt.NDArray[np.float_]
        pressure altitude at each waypoint, [:math:`Pa`]
    air_temperature : npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]
    specific_humidity: npt.NDArray[np.float_] | None
        specific humidity for each waypoint, [:math:`kg_{H_{2}O}/kg_{air}`]
    """

    if specific_humidity is None:
        specific_humidity = _estimate_specific_humidity(air_temperature, air_pressure, rh=0.6)

    # Derived quantities
    mach_num = units.tas_to_mach_number(true_airspeed, air_temperature)
    theta_amb = jet.temperature_ratio(air_temperature)
    delta_amb = jet.pressure_ratio(air_pressure)
    fuel_flow_sl = jet.equivalent_fuel_flow_rate_at_sea_level(
        fuel_flow_per_engine, theta_amb, delta_amb, mach_num
    )
    ei_nox_sl = log_ei_nox_profile.log_interp(fuel_flow_sl)

    q_correction = _get_humidity_correction_factor(specific_humidity)
    ei_cruise = ei_at_cruise(ei_nox_sl, theta_amb, delta_amb, "NOX")
    return ei_cruise * q_correction


def estimate_ei(
    log_ei_profile: EmissionsProfileInterpolator,
    fuel_flow_per_engine: npt.NDArray[np.float_],
    true_airspeed: npt.NDArray[np.float_],
    air_pressure: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """Estimate carbon monoxide (CO) or hydrocarbon (HC) emissions index (EI).

    Parameters
    ----------
    log_ei_profile: EmissionsProfileInterpolator
        emissions profile containing the log of EI CO or EI HC versus log of fuel flow.
    fuel_flow_per_engine: npt.NDArray[np.float_]
        fuel mass flow rate per engine, [:math:`kg s^{-1}`]
    true_airspeed: npt.NDArray[np.float_]
        true airspeed for each waypoint, [:math:`m s^{-1}`]
    air_pressure : npt.NDArray[np.float_]
        pressure altitude at each waypoint, [:math:`Pa`]
    air_temperature : npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]
    """
    mach_num = units.tas_to_mach_number(true_airspeed, air_temperature)
    theta_amb = jet.temperature_ratio(air_temperature)
    delta_amb = jet.pressure_ratio(air_pressure)
    fuel_flow_sl = jet.equivalent_fuel_flow_rate_at_sea_level(
        fuel_flow_per_engine, theta_amb, delta_amb, mach_num
    )

    ei_sl = log_ei_profile.log_interp(fuel_flow_sl)
    return ei_at_cruise(ei_sl, theta_amb, delta_amb, "HC")


# -----------------------
# Common helper functions
# -----------------------


def ei_at_cruise(
    ei_sl: npt.NDArray[np.float_],
    theta_amb: npt.NDArray[np.float_],
    delta_amb: npt.NDArray[np.float_],
    ei_type: str,
) -> npt.NDArray[np.float_]:
    """Convert the estimated EI at sea level to cruise conditions.

    Refer to Eqs. (15) and (16) in DuBois & Paynter (2006).

    Parameters
    ----------
    ei_sl : npt.NDArray[np.float_]
        Sea level EI values.
    theta_amb : npt.NDArray[np.float_]
        Ratio of the ambient temperature to the temperature at mean sea-level.
    delta_amb : npt.NDArray[np.float_]
        Ratio of the pressure altitude to the surface pressure.
    ei_type : str
        One of {"HC", "CO", "NOX"}

    Returns
    -------
    npt.NDArray[np.float_]
        Estimated cruise EI values.

    References
    ----------
    - :cite:`duboisFuelFlowMethod22006`
    """
    if ei_type in ["HC", "CO"]:
        # bottom of page 3, x = 1
        return ei_sl * (theta_amb**3.3 / delta_amb**1.02)
    if ei_type == "NOX":
        # bottom of page 3, y = 0.5
        y = 0.5
        return ei_sl * (delta_amb**1.02 / theta_amb**3.3) ** y
    raise ValueError("Expect ei_type to be one of 'HC', 'CO', or 'NOX'")


def _get_humidity_correction_factor(
    specific_humidity: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    return np.exp(-19 * (specific_humidity - 0.00634))


def _estimate_specific_humidity(
    air_temperature: npt.NDArray[np.float_], air_pressure: npt.NDArray[np.float_], rh: float
) -> npt.NDArray[np.float_]:
    """Estimate the specific humidity by assuming a fixed relative humidity.

    Refer to Eqs. (43), (44) and (45) in DuBois & Paynter (2006).

    Parameters
    ----------
    air_temperature : npt.NDArray[np.float_]
        Air temperature, [:math:`K`]
    air_pressure : npt.NDArray[np.float_]
        Air pressure, [:math:`Pa`]
    rh : float
        Relative humidity, [:math:`0 - 1`]

    Returns
    -------
    npt.NDArray[np.float_]
        Estimated specific humidity, [:math:`kg kg^{-1}`]

    References
    ----------
    - :cite:`duboisFuelFlowMethod22006`
    """
    # Equation (43)
    air_temperature_celcius = units.kelvin_to_celcius(air_temperature)

    # Equation (44)
    exponent = (7.5 * air_temperature_celcius) / (237.3 + air_temperature_celcius)
    P_sat = 6.107 * 10**exponent

    # Equation (45)
    air_pressure_hpa = air_pressure / 100
    numer = 0.62197058 * rh * P_sat
    denom = air_pressure_hpa - rh * P_sat
    return numer / denom



"""Cocip flight summary statistics."""

from __future__ import annotations

import numpy as np
import pandas as pd

from pycontrails.core.flight import Flight
from pycontrails.physics import units


def _get_total_fuel_consumption(fuel_flow: np.ndarray, time: np.ndarray) -> float:
    """
    Calculate the total fuel consumption of the flight.

    Parameters
    ----------
    fuel_flow : np.ndarray
        fuel mass flow rate [:math:`kg s^{-1}`]
    time : np.ndarray
        timestamp for each waypoint in np.datetime64 for each element.

    Returns
    -------
    float
        Total fuel consumption for the flight [:math:`kg`]
    """
    dt_sec = np.diff(time, append=np.datetime64("NaT")) / np.timedelta64(1, "s")
    return np.nansum(fuel_flow * dt_sec)


def _check_for_initial_contrails(sac: np.ndarray) -> bool:
    """Check if any flight waypoints satisfy the Schmidt-Appleman Criterion.

    Parameters
    ----------
    sac: np.ndarray
        Schmidt-Appleman Criterion (1: satisfies sac, 0: does not satisfy sac,
        nan: outside met domain)

    Returns
    -------
    bool
        Some or all waypoints in this flight forms contrails.
    """
    return (sac == 1).any()


def _get_mean_spatiotemporal_value(contrail: pd.DataFrame, var_name: str) -> np.float64:
    """
    Calculate the mean value of the selected contrail property.

    First, the mean value for each contrail segment is calculated over time, followed by
    taking the mean of means for each waypoint.

    Parameters
    ----------
    contrail: pd.DataFrame
        Contrail properties for each segment and its evolution
    var_name: str
        Variable name of the contrail property in contrail

    Returns
    -------
     np.float64
        Mean value of the selected contrail property
    """
    try:
        df_var = contrail[["waypoint", var_name]]
    except KeyError:
        raise KeyError(f"Variable name {var_name} is not found in the DataFrame.")
    else:
        df_var.set_index("waypoint", inplace=True, drop=True)
        df_wypt_var = df_var.groupby(df_var.index).mean()
        return np.nanmean(df_wypt_var[var_name])


def _get_mean_contrail_age(contrail: pd.DataFrame) -> np.float64:
    """
    Calculate the mean contrail age for the flight.

    This is calculated by taking the mean of the maximum contrail age for each waypoint.

    Parameters
    ----------
    contrail: pd.DataFrame
        Contrail properties for each segment and its evolution

    Returns
    -------
     np.float64
        Mean contrail age for the flight [:math:`hours`]
    """
    df_age = contrail[["waypoint", "age_hours"]]
    df_age.set_index("waypoint", inplace=True, drop=True)
    df_wypt_age_max = df_age.groupby(df_age.index).max()
    return np.nanmean(df_wypt_age_max["age_hours"])


def flight_statistics(flight: Flight, contrail: pd.DataFrame | None) -> pd.Series:
    """
    Calculate/aggregate the flight, emissions and contrail statistics for one flight.

    Parameters
    ----------
    flight : Flight
        Flight object from pycontrails.core.flight
    contrail : pd.DataFrame | None
        Contrail properties for each segment and its evolution

    Returns
    -------
    pd.Series
        Flight statistics as :class:`pandas.Series`
    """

    has_short_lived_contrails = _check_for_initial_contrails(flight["sac"])
    has_persistent_contrails = contrail is not None

    flight_stats = {
        "Flight ID": flight.attrs["flight_id"],
        "ATYP": flight.attrs["aircraft_type"],
        "First wypt": flight["time"].min(),
        "Last wypt": flight["time"].max(),
        # Fuel burn and emissions
        "Total flight dist (km)": np.nansum(flight["segment_length"]) / 1000,
        "Total fuel burn (kg)": _get_total_fuel_consumption(flight["fuel_flow"], flight["time"]),
        "OPE mean": np.nanmean(flight["engine_efficiency"]),
        "Fuel methodology": flight.attrs.get("bada_model", None),
        "ATYP assumed": flight.attrs.get("aircraft_type_bada", None),
        "Engine name": flight.attrs.get("engine_name", None),
        "Aircraft mass mean (kg)": np.nanmean(flight["aircraft_mass"]),
        # 'Total NOx emissions (kg)': self.tot_nox_emissions_kg,
        # 'NOx methodology': self.nox_methodology,
        "BC EI_n mean (kg-1)": np.nanmean(flight.get_data_or_attr("nvpm_ei_n")),
        "BC methodology": (
            flight.attrs["bc_data_source"] if "bc_data_source" in flight.attrs else None
        ),
        # Initial contrail length
        "Short-lived contrails": has_short_lived_contrails,
        "Initial contrail length (km)": (
            np.nansum(flight["segment_length"][flight["sac"] == 1]) / 1000
            if has_short_lived_contrails
            else 0
        ),
        # Persistent contrail properties
        "Persistent contrails": has_persistent_contrails,
        "First contrail wypt": contrail["time"].min() if has_persistent_contrails else "NaT",  # type: ignore  # noqa: E501
        "Last contrail wypt": contrail["time"].max() if has_persistent_contrails else "NaT",  # type: ignore  # noqa: E501
        "RHi initial, Mean": np.nanmean(flight["rhi_1"]) if has_persistent_contrails else 0,
        "RHi initial, Stdev": np.nanstd(flight["rhi_1"]) if has_persistent_contrails else 0,
        "RHi lifetime, Mean": (
            _get_mean_spatiotemporal_value(contrail, "rhi") if has_persistent_contrails else 0
        ),
        "Temp initial, Mean (K)": (
            np.nanmean(flight["air_temperature_1"]) if has_persistent_contrails else 0
        ),
        "Temp initial, Stdev (K)": (
            np.nanstd(flight["air_temperature_1"]) if has_persistent_contrails else 0
        ),
        "Temp SAC, Mean (K)": (
            np.nanmean(flight["T_critical_sac"]) if has_persistent_contrails else 0
        ),  # TODO: bugs to be resolved?
        "Temp SAC, Stdev (K)": (
            np.nanstd(flight["T_critical_sac"]) if has_persistent_contrails else 0
        ),  # TODO: bugs to be resolved?
        "Persistent contrail length (km)": (
            np.nansum(flight["segment_length"][flight["persistent_1"] == 1]) / 1000
            if has_persistent_contrails
            else 0
        ),
        "Contrail altitude initial, Mean (ft)": (
            units.m_to_ft(np.nanmean(flight["altitude_1"])) if has_persistent_contrails else 0
        ),
        "Contrail altitude lifetime, Mean (ft)": (
            units.m_to_ft(_get_mean_spatiotemporal_value(contrail, "altitude"))
            if has_persistent_contrails
            else 0
        ),
        "Contrail age, Mean (h)": (
            _get_mean_contrail_age(contrail) if has_persistent_contrails else 0
        ),
        "Contrail age, Max (h)": (
            np.nanmax(contrail["age_hours"]) if has_persistent_contrails else 0  # type: ignore
        ),
        "Ice number initial, Mean (m-1)": (
            np.nanmean(flight["n_ice_per_m_1"]) if has_persistent_contrails else 0
        ),
        "Ice number initial, Stdev (m-1)": (
            np.nanstd(flight["n_ice_per_m_1"]) if has_persistent_contrails else 0
        ),
        "Ice number lifetime, Mean (m-1)": (
            _get_mean_spatiotemporal_value(contrail, "n_ice_per_m")
            if has_persistent_contrails
            else 0
        ),
        "Ice vol mean radius, Mean (um)": (
            _get_mean_spatiotemporal_value(contrail, "r_ice_vol") * 1e6
            if has_persistent_contrails
            else 0
        ),
        "Tau contrail, Mean": (
            _get_mean_spatiotemporal_value(contrail, "tau_contrail")
            if has_persistent_contrails
            else 0
        ),
        "Tau cirrus, Mean": (
            _get_mean_spatiotemporal_value(contrail, "tau_cirrus")
            if has_persistent_contrails
            else 0
        ),
        # Radiative properties
        "RF SW (W m-2)": (
            _get_mean_spatiotemporal_value(contrail, "rf_sw") if has_persistent_contrails else 0
        ),
        "RF LW (W m-2)": (
            _get_mean_spatiotemporal_value(contrail, "rf_lw") if has_persistent_contrails else 0
        ),
        "RF Net (W m-2)": (
            _get_mean_spatiotemporal_value(contrail, "rf_net") if has_persistent_contrails else 0
        ),
        "Total contrail EF (J)": np.nansum(contrail["ef"]) if has_persistent_contrails else 0,  # type: ignore  # noqa: E501
        "SDR mean (W m-2)": (
            _get_mean_spatiotemporal_value(contrail, "sdr") if has_persistent_contrails else 0
        ),
        "RSR mean (W m-2)": (
            _get_mean_spatiotemporal_value(contrail, "rsr") if has_persistent_contrails else 0
        ),
        "OLR mean (W m-2)": (
            _get_mean_spatiotemporal_value(contrail, "olr") if has_persistent_contrails else 0
        ),
    }

    return pd.Series(flight_stats)



"""Calculate tau cirrus on Met data."""

import numpy as np
import xarray as xr

from pycontrails.core.met import MetDataset
from pycontrails.core.met_var import MetVariable
from pycontrails.physics import constants, thermo
from pycontrails.utils.types import ArrayLike

TauCirrus = MetVariable(
    short_name="tau_cirrus",
    standard_name="tau_cirrus",
    long_name="Cirrus optical depth",
    units="dimensionless",
)


def tau_cirrus(met: MetDataset) -> xr.DataArray:
    """Calculate the optical depth of NWP cirrus around each pressure level.

    Parameters
    ----------
    met : MetDataset
        A MetDataset with the following variables:
        - "air_temperature"
        - "specific_cloud_ice_water_content" or "ice_water_mixing_ratio"
        - "geopotential_height" or "geopotential"

    Returns
    -------
    xr.DataArray
        Array of tau cirrus values. Has the same dimensions as the input data.

    Notes
    -----
    Implementation differs from original Fortran implementation in computing the
    vertical derivative of geopotential height. In particular, the finite difference
    at the top-most and bottom-most layers different from original calculation by a
    factor of two. The implementation here is consistent with a numerical approximation
    of the derivative.
    """

    try:
        geopotential_height = met.data["geopotential_height"]
    except KeyError:
        geopotential_height = met.data["geopotential"] / constants.g

    # TODO: these are not *quite* the same, though we treat them the same for now
    # ECMWF "specific_cloud_ice_water_content" is mass ice per mass of *moist* air
    # GFS "ice_water_mixing_ratio" is mass ice per mass of *dry* air
    #
    # The method `cirrus_effective_extinction_coef` uses input of mass ice per mass of *dry* air,
    # so the ECMWF data is not quite right.
    try:
        ciwc = met.data["specific_cloud_ice_water_content"]
    except KeyError:
        ciwc = met.data["ice_water_mixing_ratio"]

    beta_e = cirrus_effective_extinction_coef(
        ciwc,
        met.data["air_temperature"],
        met.data["air_pressure"],
    )

    dz = -np.gradient(geopotential_height, axis=geopotential_height.get_axis_num("level"))
    dz = xr.DataArray(dz, dims=geopotential_height.dims)

    da = beta_e * dz

    da = da.cumsum(dim="level")

    return _assign_attrs(da)


def cirrus_effective_extinction_coef(
    ciwc: ArrayLike,
    T: ArrayLike,
    p: ArrayLike,
) -> ArrayLike:
    r"""Calculate the effective extinction coefficient for spectral range 0.2-0.69 um.

    Parameters
    ----------
    ciwc : ArrayLike
        Cloud ice water content, [:math:`kg_{ice} kg_{dry \ air}^{-1}`].
        Note that ECMWF provides specific ice water content per mass *moist* air.
    T : ArrayLike
        Air temperature, [:math:`K`]
    p : ArrayLike
        Air pressure, [:math:`Pa`]

    Returns
    -------
    ArrayLike
        Effective extinction coefficient for spectral range 0.2-0.69 um, [:math:`m^{-1}`]

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    - :cite:`sunParametrizationEffectiveSizes1999`

    Notes
    -----
    References as noted in :cite:`schumannContrailCirrusPrediction2012`:

    - Sun and Rikus QJRMS (1999), 125, 3037-3055
    - Sun QJRMS (2001), 127, 267-271
    - McFarquhar QJRMS (2001), 127, 261-265
    """
    # ciwc has some negative values
    # these become NaN when we raise them to powers down below
    # explicitly clipping at 0
    ciwc = ciwc.clip(min=0.0)

    # Coefficients to calculate beta_e
    a_0_beta = -1.30817e-4
    a_1_beta = 2.52883e0

    rho_air = thermo.rho_d(T, p)
    riwc = ciwc * rho_air * 1000.0

    # calculates the ice particle effective diameter in the NWP cirrus (in um)
    # Computing these exponentials is expensive. If this is a bottleneck, numexpr
    # should be used.
    tiwc = T + constants.absolute_zero + 190.0
    d_eff = 45.8966 * riwc**0.2214 + 0.7957 * tiwc * riwc**0.2535

    # explicitly clipping at 10
    d_eff = d_eff.clip(min=10.0)

    return riwc * (a_0_beta + a_1_beta / d_eff)


def _assign_attrs(da: xr.DataArray) -> xr.DataArray:
    """Assign name and attrs to xr.DataArray.

    Parameters
    ----------
    da : xr.DataArray
        DataArray to assign attributes to

    Returns
    -------
    xr.DataArray
        DataArray with assigned attributes
    """
    da.name = TauCirrus.standard_name
    da.attrs = TauCirrus.attrs

    return da



"""Pycontrails Caching Support."""

from __future__ import annotations

import logging
import os
import pathlib
import shutil
import warnings
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Sequence

logger = logging.getLogger(__name__)

from overrides import overrides

# optional imports
if TYPE_CHECKING:
    import google


def _get_user_cache_dir() -> str:
    try:
        import platformdirs
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            "Using the pycontrails CacheStore requires the 'platformdirs' package. "
            "This can be installed with 'pip install pycontrails[ecmwf]' or "
            "'pip install platformdirs'."
        ) from e
    return platformdirs.user_cache_dir("pycontrails")


class CacheStore(ABC):
    """Abstract cache storage class for storing staged and intermediate data."""

    __slots__ = ("cache_dir", "allow_clear")
    cache_dir: str
    allow_clear: bool

    @property
    @abstractmethod
    def size(self) -> float:
        """Return the disk size (in MBytes) of the local cache.

        Returns
        -------
        float
            Size of the disk cache store in MB

        Examples
        --------
        >>> from pycontrails import DiskCacheStore
        >>> disk_cache = DiskCacheStore(cache_dir="cache", allow_clear=True)
        >>> disk_cache.size
        0.0...

        >>> disk_cache.clear()  # cleanup

        >>> from pycontrails import GCPCacheStore
        >>> gcp_cache = GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="cache")
        >>> gcp_cache.size
        0.0...

        """

    @abstractmethod
    def listdir(self, path: str = "") -> list[str]:
        """List the contents of a directory in the cache.

        Parameters
        ----------
        path : str
            Path to the directory to list

        Returns
        -------
        list[str]
            List of files in the directory
        """

    @abstractmethod
    def path(self, cache_path: str) -> str:
        """Return a full filepath in cache.

        Parameters
        ----------
        cache_path : str
            string path or filepath to create in cache
            If parent directories do not exist, they will be created.

        Returns
        -------
        str
            Full path string to subdirectory directory or object in cache directory

        Examples
        --------
        >>> from pycontrails import DiskCacheStore
        >>> disk_cache = DiskCacheStore(cache_dir="cache", allow_clear=True)
        >>> disk_cache.path("file.nc")
        'cache/file.nc'

        >>> disk_cache.clear()  # cleanup

        >>> from pycontrails import GCPCacheStore
        >>> gcp_cache = GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="cache")
        >>> gcp_cache.path("file.nc")
        'cache/file.nc'
        """

    @abstractmethod
    def exists(self, cache_path: str) -> bool:
        """Check if a path in cache exists.

        Parameters
        ----------
        cache_path : str
            Path to directory or file in cache

        Returns
        -------
        bool
            True if directory or file exists

        Examples
        --------
        >>> from pycontrails import DiskCacheStore
        >>> disk_cache = DiskCacheStore(cache_dir="cache", allow_clear=True)
        >>> disk_cache.exists("file.nc")
        False

        >>> from pycontrails import GCPCacheStore
        >>> gcp_cache = GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="cache")
        >>> gcp_cache.exists("file.nc")
        False
        """

    def put_multiple(
        self, data_path: Sequence[str | pathlib.Path], cache_path: list[str]
    ) -> list[str]:
        """Put multiple files into the cache at once.

        Parameters
        ----------
        data_path : Sequence[str | pathlib.Path]
            List of data files to cache.
            Each member is passed directly on to :meth:`put`.
        cache_path : list[str]
            List of cache paths corresponding to each element in the ``data_path`` list.
            Each member is passed directly on to :meth:`put`.

        Returns
        -------
        list[str]
            Returns a list of relative paths to the stored files in the cache
        """

        # TODO: run in parallel?
        return [self.put(d, cp) for d, cp in zip(data_path, cache_path)]

    # In the three methods below, child classes have a complete docstring.

    @abstractmethod
    def put(self, data: str | pathlib.Path, cache_path: str | None = None) -> str:
        """Save data to cache."""

    @abstractmethod
    def get(self, cache_path: str) -> str:
        """Get data from cache."""


class DiskCacheStore(CacheStore):
    """Cache that uses a folder on the local filesystem.

    Parameters
    ----------
    allow_clear : bool, optional
        Allow this cache to be cleared using :meth:`clear()`. Defaults to False.
    cache_dir : str | pathlib.Path, optional
        Root cache directory.
        By default, looks first for ``PYCONTRAILS_CACHE_DIR`` environment variable,
        then uses the OS specific :func:`platformdirs.user_cache_dir` function.

    Examples
    --------
    >>> from pycontrails import DiskCacheStore
    >>> disk_cache = DiskCacheStore(cache_dir="cache", allow_clear=True)
    >>> disk_cache.cache_dir
    'cache'

    >>> disk_cache.clear()  # cleanup
    """

    def __init__(
        self,
        cache_dir: str | pathlib.Path | None = None,
        allow_clear: bool = False,
    ):
        if cache_dir is None:
            # Use a try / except to avoid unnecessary import of platformdirs
            try:
                cache_dir = os.environ["PYCONTRAILS_CACHE_DIR"]
            except KeyError:
                cache_dir = _get_user_cache_dir()

        # make sure local cache directory exists
        pathlib.Path(cache_dir).mkdir(parents=True, exist_ok=True)

        # store root cache dir
        self.cache_dir = str(cache_dir)

        # allow the cache to be clear or not
        self.allow_clear = allow_clear

    def __repr__(self) -> str:
        return f"DiskCacheStore: {self.cache_dir}"

    @property
    @overrides
    def size(self) -> float:
        disk_path = pathlib.Path(self.cache_dir)
        size = sum(f.stat().st_size for f in disk_path.rglob("*") if f.is_file())
        logger.debug("Disk cache size %s bytes", size)
        return size / 1e6

    @overrides
    def listdir(self, path: str = "") -> list[str]:
        path = self.path(path)
        iter_ = pathlib.Path(path).iterdir()
        out = [str(f.relative_to(path)) for f in iter_]
        out.sort()
        return out

    @overrides
    def path(self, cache_path: str) -> str:
        if str(cache_path).startswith(str(self.cache_dir)):
            disk_path = pathlib.Path(cache_path)
        else:
            disk_path = pathlib.Path(self.cache_dir) / pathlib.Path(cache_path)

        # make sure full path to parents exist
        disk_path.parent.mkdir(parents=True, exist_ok=True)

        return str(disk_path)

    @overrides
    def exists(self, cache_path: str) -> bool:
        disk_path = pathlib.Path(self.path(cache_path))
        return disk_path.is_dir() or disk_path.is_file()

    def put(self, data_path: str | pathlib.Path, cache_path: str | None = None) -> str:
        """Save data to the local cache store.

        Parameters
        ----------
        data_path : str | pathlib.Path
            Path to data to cache.
        cache_path : str | None, optional
            Path in cache store to save data
            Defaults to the same filename as ``data_path``

        Returns
        -------
        str
            Returns the relative path in the cache to the stored file

        Raises
        ------
        FileNotFoundError
            Raises if `data` is a string and a file is not found at the string

        Examples
        --------
        >>> from pycontrails import DiskCacheStore
        >>> disk_cache = DiskCacheStore(cache_dir="cache", allow_clear=True)
        >>>
        >>> # put a file directly
        >>> disk_cache.put("README.md", "test/file.md")
        'test/file.md'
        """

        if not pathlib.Path(data_path).is_file():
            raise FileNotFoundError(f"No file found at path {data_path}")

        if cache_path is None:
            cache_path = pathlib.Path(data_path).name

        disk_path = self.path(str(cache_path))

        # copy to disk cache
        logger.debug("Disk cache put %s to %s in disk cache", data_path, cache_path)
        try:
            shutil.copyfile(data_path, disk_path)
        except PermissionError:
            logger.warning(
                "Permission error copying %s to %s. The destination file may already be open.",
                data_path,
                disk_path,
            )

        return cache_path

    def get(self, cache_path: str) -> str:
        """Get data path from the local cache store.

        Alias for :meth:`path()`

        Parameters
        ----------
        cache_path : str
            Cache path to retrieve

        Returns
        -------
        str
            Returns the relative path in the cache to the stored file

        Examples
        --------
        >>> from pycontrails import DiskCacheStore
        >>> disk_cache = DiskCacheStore(cache_dir="cache", allow_clear=True)
        >>>
        >>> # returns a path
        >>> disk_cache.get("test/file.md")
        'cache/test/file.md'
        """
        return self.path(cache_path)

    def clear(self, cache_path: str = "") -> None:
        """Delete all files and folders within ``cache_path``.

        If no ``cache_path`` is provided, this will clear the entire cache.

        If :attr:`allow_clear` is set to ``False``, this method will do nothing.

        Parameters
        ----------
        cache_path : str, optional
            Path to subdirectory or file in cache

        Raises
        ------
        RuntimeError
            Raises a RuntimeError when :attr:`allow_clear` is set to ``False``

        Examples
        --------
        >>> from pycontrails import DiskCacheStore
        >>> disk_cache = DiskCacheStore(cache_dir="cache", allow_clear=True)

        >>> # Write some data to the cache
        >>> disk_cache.put("README.md", "test/example.txt")
        'test/example.txt'

        >>> disk_cache.exists("test/example.txt")
        True

        >>> # clear a specific path
        >>> disk_cache.clear("test/example.txt")

        >>> # clear the whole cache
        >>> disk_cache.clear()
        """
        if not self.allow_clear:
            raise RuntimeError("Cache is not allowed to be cleared")

        disk_path = pathlib.Path(self.path(cache_path))

        if disk_path.is_file():
            logger.debug("Remove file at path %s", disk_path)
            disk_path.unlink()

        # Assume anything else is a directory
        elif disk_path.exists():
            # rm directory recursively
            logger.debug("Remove directory at path %s", disk_path)
            shutil.rmtree(disk_path, ignore_errors=True)

        else:
            warnings.warn(f"No cache path found at {disk_path}")

        # make sure local cache directory exists
        # pathlib.Path(self.cache_dir).mkdir(parents=True, exist_ok=True)


class GCPCacheStore(CacheStore):
    """Google Cloud Platform (Storage) Cache.

    This class downloads files from Google Cloud Storage locally to a :class:`DiskCacheStore`
    initialized with ``cache_dir=".gcp"`` to avoid re-downloading files. If the source files
    on GCP changes, the local mirror of the GCP DiskCacheStore must be cleared by initializing
    this class and running :meth:`clear_disk()`.

    Note by default, GCP Cache Store is *read only*.
    When a :meth:`put` is called and :attr:`read_only` is set to *True*,
    the cache will throw an ``RuntimeError`` error.
    Set ``read_only`` to *False* to enable writing to cache store.

    Parameters
    ----------
    cache_dir : str, optional
        Root object prefix within :attr:`bucket`
        Defaults to ``PYCONTRAILS_CACHE_DIR`` environment variable, or the root of the bucket.
        The full GCP URI (ie, `"gs://<MY_BUCKET>/<PREFIX>"`) can be used here.
    project : str , optional
        GCP Project.
        Defaults to the current active project set in the `google-cloud-sdk` environment
    bucket : str, optional
        GCP Bucket to use for cache.
        Defaults to ``PYCONTRAILS_CACHE_BUCKET`` environment variable.
    read_only : bool, optional
        Only enable reading from cache. Defaults to ``True``.
    allow_clear : bool, optional
        Allow this cache to be cleared using :meth:`clear()`. Defaults to ``False``.
    disk_cache : DiskCacheStore, optional
        Specify a custom local disk cache store to mirror files.
        Defaults to :class:`DiskCacheStore(cache_dir="{user_cache_dir}/.gcp/{bucket}")`
    show_progress : bool, optional
        Show progress bar on cache :meth:`put`.
        Defaults to False
    chunk_size : int, optional
        Chunk size for uploads and downloads with progress. Set a larger size to see more granular
        progress, and set a smaller size for more optimal download speed. Chunk size must be a
        multiple of 262144 (ie, 10 * 262144). Default value is 8 * 262144, which will throttle
        fast download speeds.


    Examples
    --------
    >>> from pycontrails import GCPCacheStore
    >>> cache = GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="cache")
    >>> cache.cache_dir
    'cache/'
    >>> cache.bucket
    'contrails-301217-unit-test'
    """

    __slots__ = (
        "project",
        "bucket",
        "read_only",
        "timeout",
        "show_progress",
        "chunk_size",
        "_disk_cache",
        "_client",
        "_bucket",
    )
    project: str | None
    bucket: str
    read_only: bool
    timeout: int
    show_progress: bool
    chunk_size: int
    _disk_cache: DiskCacheStore
    _client: "google.cloud.storage.Client"
    _bucket: "google.cloud.storage.Bucket"

    def __init__(
        self,
        cache_dir: str = os.getenv("PYCONTRAILS_CACHE_DIR", ""),
        project: str | None = None,
        bucket: str | None = os.getenv("PYCONTRAILS_CACHE_BUCKET"),
        disk_cache: DiskCacheStore | None = None,
        read_only: bool = True,
        allow_clear: bool = False,
        timeout: int = 300,
        show_progress: bool = False,
        chunk_size: int = 64 * 262144,
    ):
        try:
            from google.cloud import storage
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "GCPCache requires the `google-cloud-storage` module, which can be installed "
                "using `pip install pycontrails[gcp]`"
            ) from e

        if "https://" in cache_dir:
            raise ValueError(
                "`cache_dir` should only specify base object path within the GCS bucket. "
                "Expect not to find prefix `https://` in parameter `cache_dir`. "
                f"Found `cache_dir={cache_dir}`."
            )

        # support cache_dir paths that refer to the whole GCP URI path
        if "gs://" in cache_dir:
            bucket_and_cache_dir = cache_dir.split("gs://")[1]
            split_path = bucket_and_cache_dir.split("/", maxsplit=1)
            if len(split_path) == 1:
                uri_bucket = split_path[0]
                cache_dir = ""
            else:
                uri_bucket, cache_dir = split_path

            if bucket is None:
                bucket = uri_bucket
            elif bucket != uri_bucket:
                raise ValueError(
                    f"Found conflicting bucket names: {uri_bucket} in URI path "
                    f"and {bucket} as parameter."
                )

        # TODO: Not sure if we want this ....
        # Do we want to correct for parameters bucket=None and cache_dir=BUCKET/PREFIX?
        # if bucket in cache_dir:
        #     cache_dir = cache_dir.split(f"{bucket}/")[1]

        # raise if bucket is still not defined
        if bucket is None:
            raise ValueError(
                "Parameter `bucket` not specified. Either pass parameter `bucket`, pass a URI "
                "path for `cache_dir`, or set environment variable `PYCONTRAILS_CACHE_BUCKET`"
            )

        # append a "/" for GCP objects
        if cache_dir != "" and not cache_dir.endswith("/"):
            cache_dir = f"{cache_dir}/"

        # set up gcp client
        self._client = storage.Client(project=project)

        # create bucket object and make sure bucket exists
        self._bucket = self._client.bucket(bucket)

        # store root bucket/cache dir
        self.project = project
        self.bucket = bucket
        self.cache_dir = cache_dir

        # read only
        self.read_only = read_only

        # allow the cache to be cleared or not
        self.allow_clear = allow_clear

        # parameters for GCP storage upload
        self.timeout = timeout
        self.show_progress = show_progress
        self.chunk_size = chunk_size

        # set up local DiskCache mirror
        # this keeps a local copy of files so that files are not re-downloaded
        if disk_cache is not None:
            self._disk_cache = disk_cache
        else:
            local_cache_dir = _get_user_cache_dir()
            self._disk_cache = DiskCacheStore(
                cache_dir=f"{local_cache_dir}/.gcp/{bucket}", allow_clear=True
            )

    def __repr__(self) -> str:
        return f"GCPCacheStore: {self.bucket}/{self.cache_dir}"

    @property
    def client(self) -> "google.cloud.storage.Client":
        """Handle to Google Cloud Storage client.

        Returns
        -------
        :class:`google.cloud.storage.Client`
            Handle to Google Cloud Storage client
        """
        return self._client

    @property
    @overrides
    def size(self) -> float:
        # get list of blobs below this path
        blobs = self._bucket.list_blobs(prefix=self.cache_dir)
        size = sum(b.size for b in blobs)
        logger.debug("GCP cache size %s bytes", size)
        return size / 1e6

    @overrides
    def listdir(self, path: str = "") -> list[str]:
        # I don't necessarily think we want to implement this .... it might be
        # very slow if the bucket is large. BUT, it won't be slower than the size
        # method right above this.
        # I typically am more interested in calling self._disk_cache.listdir() to get
        # information about the local cache, which is why I include this
        # particular error message.
        raise NotImplementedError(
            "ls is not implemented for GCPCacheStore. Use ._disk_cache.listdir() to "
            "list files in the local disk cache."
        )

    @overrides
    def path(self, cache_path: str) -> str:
        if cache_path.startswith(self.cache_dir):
            return cache_path
        return f"{self.cache_dir}{cache_path}"

    def gs_path(self, cache_path: str) -> str:
        """Return a full Google Storage (gs://) URI to object.

        Parameters
        ----------
        cache_path : str
            string path to object in cache

        Returns
        -------
        str
            Google Storage URI (gs://) to object in cache

        Examples
        --------
        >>> from pycontrails import GCPCacheStore
        >>> cache = GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="cache")
        >>> cache.path("file.nc")
        'cache/file.nc'
        """
        bucket_path = self.path(cache_path)
        return f"gs://{self.bucket}/{bucket_path}"

    @overrides
    def exists(self, cache_path: str) -> bool:
        # see if file is in the mirror disk cache
        if self._disk_cache.exists(cache_path):
            return self._disk_cache.exists(cache_path)

        bucket_path = self.path(cache_path)
        blob = self._bucket.blob(bucket_path)

        return blob.exists()

    def put(self, data_path: str | pathlib.Path, cache_path: str | None = None) -> str:
        """Save data to the GCP cache store.

        If :attr:`read_only` is *True*, this method will return the path to the
        local disk cache store path.

        Parameters
        ----------
        data_path : str | pathlib.Path
            Data to save to GCP cache store.
        cache_path : str, optional
            Path in cache store to save data.
            Defaults to the same filename as ``data_path``.

        Returns
        -------
        str
            Returns the path in the cache to the stored file

        Raises
        ------
        RuntimeError
            Raises if :attr:`read_only` is True
        FileNotFoundError
            Raises if ``data`` is a string and a file is not found at the string

        Examples
        --------
        >>> from pycontrails import GCPCacheStore
        >>> cache = GCPCacheStore(
        ...     bucket="contrails-301217-unit-test",
        ...     cache_dir="cache",
        ...     read_only=False,
        ... )

        >>> # put a file directly
        >>> cache.put("README.md", "test/file.md")
        'test/file.md'
        """
        # store on disk path mirror -  will catch errors
        cache_path = self._disk_cache.put(data_path, cache_path)

        # read only
        if self.read_only:
            logger.debug(
                f"GCP Cache Store is read only. File put in local DiskCacheStore path: {cache_path}"
            )
            raise RuntimeError(
                f"GCP Cache Store {self.bucket}/{self.cache_dir} is read only. "
                "File put in local DiskCacheStore path: {cache_path}"
            )

        # get bucket and disk paths and blob
        bucket_path = self.path(cache_path)
        disk_path = self._disk_cache.path(cache_path)
        blob = self._bucket.blob(bucket_path)

        logger.debug("GCP Cache put %s to %s", disk_path, bucket_path)

        if self.show_progress:  # upload with pbar
            _upload_with_progress(blob, disk_path, self.timeout, chunk_size=self.chunk_size)
        else:  # upload from disk path
            blob.upload_from_filename(disk_path, timeout=self.timeout)

        return cache_path

    def get(self, cache_path: str) -> str:
        """Get data from the local cache store.

        Parameters
        ----------
        cache_path : str
            Path in cache store to get data

        Returns
        -------
        str
            Returns path to downloaded local file

        Raises
        ------
        ValueError
            Raises value error is ``cache_path`` refers to a directory

        Examples
        --------
        >>> import pathlib
        >>> from pycontrails import GCPCacheStore
        >>> cache = GCPCacheStore(
        ...     bucket="contrails-301217-unit-test",
        ...     cache_dir="cache",
        ...     read_only=False,
        ... )

        >>> cache.put("README.md", "example/file.md")
        'example/file.md'

        >>> # returns a full path to local copy of the file
        >>> path = cache.get("example/file.md")
        >>> pathlib.Path(path).is_file()
        True

        >>> pathlib.Path(path).read_text()[17:69]
        'Python library for modeling aviation climate impacts'
        """
        if cache_path.endswith("/"):
            raise ValueError("`cache_path` must not end with a /")

        # see if file is in the mirror disk cache
        if self._disk_cache.exists(cache_path):
            return self._disk_cache.get(cache_path)

        # download otherwise
        bucket_path = self.path(cache_path)
        disk_path = self._disk_cache.path(cache_path)

        blob = self._bucket.blob(bucket_path)
        if not blob.exists():
            raise ValueError(f"No object exists in cache at path {bucket_path}")

        logger.debug("GCP Cache GET from %s", bucket_path)

        if self.show_progress:
            _download_with_progress(
                gcp_cache=self,
                gcp_path=bucket_path,
                disk_path=disk_path,
                chunk_size=self.chunk_size,
            )
        else:
            blob.download_to_filename(disk_path)

        return self._disk_cache.get(disk_path)

    def clear_disk(self, cache_path: str = "") -> None:
        """Clear the local disk cache mirror of the GCP Cache Store.

        Parameters
        ----------
        cache_path : str, optional
            Path in mirrored cache store. Passed into :meth:`_disk_clear.clear`. By
            default, this method will clear the entire mirrored cache store.

        Examples
        --------
        >>> from pycontrails import GCPCacheStore
        >>> gcp_cache = GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="cache")
        >>> gcp_cache.clear_disk()
        """
        self._disk_cache.clear(cache_path)

    def _dangerous_clear(self, confirm: bool = False, cache_path: str = "") -> None:
        """Delete all files and folders within ``cache_path``.

        If no ``cache_path`` is provided, this will clear the entire cache.

        If :attr:`allow_clear` is set to ``False``, this method will do nothing.

        Parameters
        ----------
        confirm : bool, optional
            Must pass True to make this work
        cache_path : str, optional
            Path to sub-directory or file in cache

        Raises
        ------
        RuntimeError
            Raises a RuntimeError when :attr:`allow_clear` is set to ``False``
        """
        if not confirm or not self.allow_clear:
            raise RuntimeError("Cache is not allowed to be cleared")

        # get full path to clear
        bucket_path = self.path(cache_path)
        logger.debug("Clearing GCP cache at path %s", bucket_path)

        # clear disk mirror
        self.clear_disk()

        # get list of blobs below this path
        blobs = self._bucket.list_blobs(prefix=bucket_path)

        # clear blobs one at a time
        for blob in blobs:
            blob.delete()


def _upload_with_progress(blob: Any, disk_path: str, timeout: int, chunk_size: int) -> None:
    """Upload with `tqdm` progress bar.

    Adapted from
    https://github.com/googleapis/python-storage/issues/27#issuecomment-651468428.

    Parameters
    ----------
    blob : Any
        GCP blob to upload
    disk_path : str
        Path to local file.
    timeout : int
        Passed into `blob.upload_from_file`
    chunk_size : int
        Used to set :attr:`chunk_size` on `blob`.
    """
    try:
        from tqdm.auto import tqdm
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            "Method `put` requires the `tqdm` module, which can be installed using "
            "`pip install pycontrails[gcp]`. "
            "Alternatively, set instance attribute `show_progress=False`."
        ) from e

    # minimal possible chunk_size to allow nice progress bar
    blob.chunk_size = chunk_size

    with open(disk_path, "rb") as local_file:
        total_bytes = os.fstat(local_file.fileno()).st_size
        with tqdm.wrapattr(local_file, "read", total=total_bytes, desc="upload to GCP") as file_obj:
            blob.upload_from_file(file_obj, size=total_bytes, timeout=timeout)


def _download_with_progress(
    gcp_cache: GCPCacheStore, gcp_path: str, disk_path: str, chunk_size: int
) -> None:
    """Download with `tqdm` progress bar."""

    try:
        from google.resumable_media.requests import ChunkedDownload
        from tqdm.auto import tqdm
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            "Method `get` requires the `tqdm` and `google-cloud-storage` modules, "
            "which can be installed using `pip install pycontrails[gcp]`. "
            "Alternatively, set instance attribute `show_progress=False`."
        ) from e

    blob = gcp_cache._bucket.get_blob(gcp_path)
    url = blob._get_download_url(gcp_cache._client)
    description = f"Download {gcp_path}"

    with open(disk_path, "wb") as local_file:
        with tqdm.wrapattr(local_file, "write", total=blob.size, desc=description) as file_obj:
            download = ChunkedDownload(url, chunk_size, file_obj)
            transport = gcp_cache.client._http
            while not download.finished:
                download.consume_next_chunk(transport, timeout=gcp_cache.timeout)
